# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import math

import pandas as pd
from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis
from ax.analysis.plotly.color_constants import METRIC_CONTINUOUS_COLOR_SCALE

from ax.analysis.plotly.plotly_analysis import (
    create_plotly_analysis_card,
    PlotlyAnalysisCard,
)
from ax.analysis.plotly.surface.utils import (
    get_parameter_values,
    is_axis_log_scale,
    select_fixed_value,
)
from ax.analysis.plotly.utils import select_metric, truncate_label
from ax.analysis.utils import extract_relevant_adapter, relativize_data
from ax.core.experiment import Experiment
from ax.core.observation import ObservationFeatures
from ax.core.trial_status import STATUSES_EXPECTING_DATA
from ax.core.utils import get_target_trial_index
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from plotly import graph_objects as go
from pyre_extensions import none_throws, override

CONTOUR_CARDGROUP_TITLE = "Contour Plots: Metric effects by parameter values"

CONTOUR_CARDGROUP_SUBTITLE = (
    "These plots show the relationship between a metric and two parameters. They "
    "show the predicted values of the metric (indicated by color) as a function of "
    "the two parameters on the x- and y-axes while keeping all other parameters "
    "fixed at their status_quo value (or mean value if status_quo is unavailable). "
)


class ContourPlot(Analysis):
    """
    Plot a 2D surface of the surrogate model's predicted outcomes for a given pair of
    parameters, where all other parameters are held fixed at their status-quo value or
    mean if no status quo is available.

    The DataFrame computed will contain the following columns:
        - PARAMETER_NAME: The value of the x parameter specified
        - PARAMETER_NAME: The value of the y parameter specified
        - METRIC_NAME: The predected mean of the metric specified
        - sampled: Whether the parameter values were sampled in at least one trial
    """

    def __init__(
        self,
        x_parameter_name: str,
        y_parameter_name: str,
        metric_name: str | None = None,
        display_sampled: bool = True,
        relativize: bool = False,
    ) -> None:
        """
        Args:
            y_parameter_name: The name of the parameter to plot on the x-axis.
            y_parameter_name: The name of the parameter to plot on the y-axis.
            metric_name: The name of the metric to plot
            display_sampled: If True, plot "x"s at x coordinates which have been
                sampled in at least one trial.
            relativize: If True, relativize the metric values to the status quo.
        """
        self.x_parameter_name = x_parameter_name
        self.y_parameter_name = y_parameter_name
        self.metric_name = metric_name
        self._display_sampled = display_sampled
        self.relativize = relativize

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> PlotlyAnalysisCard:
        if experiment is None:
            raise UserInputError("ContourPlot requires an Experiment")

        relevant_adapter = extract_relevant_adapter(
            experiment=experiment,
            generation_strategy=generation_strategy,
            adapter=adapter,
        )

        metric_name = self.metric_name or select_metric(experiment=experiment)

        df = _prepare_data(
            experiment=experiment,
            model=relevant_adapter,
            x_parameter_name=self.x_parameter_name,
            y_parameter_name=self.y_parameter_name,
            metric_name=metric_name,
            relativize=self.relativize,
        )

        fig = _prepare_plot(
            df=df,
            x_parameter_name=self.x_parameter_name,
            y_parameter_name=self.y_parameter_name,
            metric_name=metric_name,
            log_x=is_axis_log_scale(
                parameter=experiment.search_space.parameters[self.x_parameter_name]
            ),
            log_y=is_axis_log_scale(
                parameter=experiment.search_space.parameters[self.y_parameter_name]
            ),
            display_sampled=self._display_sampled,
            is_relative=self.relativize,
        )

        return create_plotly_analysis_card(
            name=self.__class__.__name__,
            title=(
                f"{self.x_parameter_name}, {self.y_parameter_name} vs. "
                f"{metric_name}"
            ),
            subtitle=(
                "The contour plot visualizes the predicted outcomes "
                f"for {metric_name} across a two-dimensional parameter space, "
                "with other parameters held fixed at their status_quo value "
                "(or mean value if status_quo is unavailable). This plot helps "
                "in identifying regions of optimal performance and understanding "
                "how changes in the selected parameters influence the predicted "
                "outcomes. Contour lines represent levels of constant predicted "
                "values, providing insights into the gradient and potential optima "
                "within the parameter space."
            ),
            df=df,
            fig=fig,
        )


def compute_contour_adhoc(
    x_parameter_name: str,
    y_parameter_name: str,
    experiment: Experiment,
    generation_strategy: GenerationStrategy | None = None,
    adapter: Adapter | None = None,
    metric_name: str | None = None,
    display_sampled: bool = True,
    relativize: bool = False,
) -> PlotlyAnalysisCard:
    """
    Helper method to expose adhoc contour plotting. Only for advanced users in
    a notebook setting.

    Args:
        parameter_name: The name of the parameter to plot on the x-axis.
        experiment: The experiment to source data from.
        generation_strategy: Optional. The generation strategy to extract the adapter
            from.
        adapter: Optional. The adapter to use for predictions.
        metric_name: The name of the metric to plot on the y-axis. If not specified
            the objective will be used.
        display_sampled: If True, plot "x"s at x coordinates which have been sampled
            in at least one trial.
        relativize: If True, relativize the metric values to the status quo.
    """
    analysis = ContourPlot(
        x_parameter_name=x_parameter_name,
        y_parameter_name=y_parameter_name,
        metric_name=metric_name,
        display_sampled=display_sampled,
        relativize=relativize,
    )
    return analysis.compute(
        experiment=experiment,
        generation_strategy=generation_strategy,
        adapter=adapter,
    )


def _prepare_data(
    experiment: Experiment,
    model: Adapter,
    x_parameter_name: str,
    y_parameter_name: str,
    metric_name: str,
    relativize: bool,
) -> pd.DataFrame:
    sampled = [
        {
            "x_parameter_name": arm.parameters[x_parameter_name],
            "y_parameter_name": arm.parameters[y_parameter_name],
            "arm_name": arm.name,
            "trial_index": trial.index,
        }
        for trial in experiment.trials.values()
        if trial.status in STATUSES_EXPECTING_DATA  # running, completed, early stopped
        for arm in trial.arms
        # Filter out arms which are not part of the search space (ex. when a parameter
        # is None).
        if experiment.search_space.check_membership(
            parameterization=arm.parameters,
            raise_error=False,
            check_all_parameters_present=False,
        )
    ]

    # Choose which parameter values to predict points for.
    unsampled_xs = get_parameter_values(
        parameter=experiment.search_space.parameters[x_parameter_name], density=10
    )
    unsampled_ys = get_parameter_values(
        parameter=experiment.search_space.parameters[y_parameter_name], density=10
    )

    xs = [*[sample["x_parameter_name"] for sample in sampled], *unsampled_xs]
    ys = [*[sample["y_parameter_name"] for sample in sampled], *unsampled_ys]

    # Construct observation features for each parameter value previously chosen by
    # fixing all other parameters to their status-quo value or mean.
    features = [
        ObservationFeatures(
            parameters={
                x_parameter_name: x,
                y_parameter_name: y,
                **{
                    parameter.name: select_fixed_value(parameter=parameter)
                    for parameter in experiment.search_space.parameters.values()
                    if not (
                        parameter.name == x_parameter_name
                        or parameter.name == y_parameter_name
                    )
                },
            }
        )
        for x in xs
        for y in ys
    ]

    predictions = model.predict(observation_features=features)

    df = none_throws(
        pd.DataFrame.from_records(
            [
                {
                    x_parameter_name: features[i].parameters[x_parameter_name],
                    y_parameter_name: features[i].parameters[y_parameter_name],
                    f"{metric_name}_mean": predictions[0][metric_name][i],
                    f"{metric_name}_sem": predictions[1][metric_name][metric_name][i]
                    ** 0.5,
                    "sampled": (
                        features[i].parameters[x_parameter_name],
                        features[i].parameters[y_parameter_name],
                    )
                    in [
                        (s["x_parameter_name"], s["y_parameter_name"]) for s in sampled
                    ],
                    "arm_name": sampled[i]["arm_name"]
                    if i < len(sampled)
                    else "unsampled",
                    "trial_index": sampled[i]["trial_index"]
                    if i < len(sampled)
                    else -1,
                }
                for i in range(len(features))
            ]
        ).drop_duplicates()
    )

    if relativize:
        target_trial_index = none_throws(get_target_trial_index(experiment=experiment))
        df = relativize_data(
            experiment=experiment,
            df=df,
            metric_names=[metric_name],
            is_raw_data=False,
            trial_index=None,
            trial_statuses=None,
            target_trial_index=target_trial_index,
        )

    return df


def _prepare_plot(
    df: pd.DataFrame,
    x_parameter_name: str,
    y_parameter_name: str,
    metric_name: str,
    log_x: bool,
    log_y: bool,
    display_sampled: bool,
    is_relative: bool,
) -> go.Figure:
    z_grid = df.pivot_table(
        index=y_parameter_name,
        columns=x_parameter_name,
        values=f"{metric_name}_mean",
        # aggfunc is required to gracefully handle duplicate values
        aggfunc="mean",
    )

    if is_relative:
        z_values = z_grid.values * 100
    else:
        z_values = z_grid.values

    fig = go.Figure(
        data=go.Contour(
            z=z_values,
            x=z_grid.columns.values,
            y=z_grid.index.values,
            colorscale=METRIC_CONTINUOUS_COLOR_SCALE,
            showscale=True,
            colorbar={
                "title": None,
                "ticksuffix": "%" if is_relative else "",
            },
            hoverinfo="skip",
        ),
        layout=go.Layout(
            xaxis_title=truncate_label(label=x_parameter_name),
            yaxis_title=truncate_label(label=y_parameter_name),
        ),
    )

    if display_sampled:
        x_sampled = df[df["sampled"]][x_parameter_name].tolist()
        y_sampled = df[df["sampled"]][y_parameter_name].tolist()
        samples = go.Scatter(
            x=x_sampled,
            y=y_sampled,
            mode="markers",
            marker={
                "symbol": "x",
                "color": "black",
            },
            name="Sampled",
            showlegend=False,
            hovertemplate="(%{x}, %{y})<extra>Sampled</extra>",
        )

        fig.add_trace(samples)

    # Set the x-axis scale to log if relevant
    if log_x:
        fig.update_xaxes(
            type="log",
            range=[
                math.log10(df[x_parameter_name].min()),
                math.log10(df[x_parameter_name].max()),
            ],
        )
    else:
        fig.update_xaxes(range=[df[x_parameter_name].min(), df[x_parameter_name].max()])

    if log_y:
        fig.update_yaxes(
            type="log",
            range=[
                math.log10(df[y_parameter_name].min()),
                math.log10(df[y_parameter_name].max()),
            ],
        )
    else:
        fig.update_yaxes(range=[df[y_parameter_name].min(), df[y_parameter_name].max()])

    return fig



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import inspect
import pydoc
from collections.abc import Callable
from types import FunctionType
from typing import Any, TypeVar, Union


T = TypeVar("T")
TDecoderRegistry = dict[str, Union[type[T], Callable[..., T]]]
TClassDecoderRegistry = dict[str, Callable[[dict[str, Any]], Any]]


# https://stackoverflow.com/a/39235373
def named_tuple_to_dict(data: Any) -> Any:
    """Recursively convert NamedTuples to dictionaries."""
    if isinstance(data, dict):
        return {key: named_tuple_to_dict(value) for key, value in data.items()}
    elif isinstance(data, list):
        return [named_tuple_to_dict(value) for value in data]
    elif _is_named_tuple(data):
        return {
            key: named_tuple_to_dict(value) for key, value in data._asdict().items()
        }
    elif isinstance(data, tuple):
        return tuple(named_tuple_to_dict(value) for value in data)
    else:
        return data


# https://stackoverflow.com/a/2166841
def _is_named_tuple(x: Any) -> bool:
    """Return True if x is an instance of NamedTuple."""
    t = type(x)
    b = t.__bases__
    if len(b) != 1 or b[0] != tuple:
        return False
    f = getattr(t, "_fields", None)
    if not isinstance(f, tuple):
        return False  # pragma nocover
    return all(isinstance(n, str) for n in f)


def callable_to_reference(callable: Callable) -> str:
    """Obtains path to the callable of form <module>.<name>."""
    if not isinstance(callable, (FunctionType, type)):
        raise TypeError(f"Expected to encode function or class, got: {callable}.")
    name = f"{callable.__module__}.{callable.__qualname__}"
    try:
        assert pydoc.locate(name) is callable
        return name
    except Exception as err:
        raise TypeError(
            f"Callable {callable.__qualname__} is not properly exposed in "
            f"{callable.__module__} (exception: {err})."
        )


def callable_from_reference(path: str) -> Callable:
    """Retrieves a callable by its path."""
    return pydoc.locate(path)  # pyre-ignore[7]


def serialize_init_args(
    obj: Any,
    exclude_fields: list[str] | None = None,
) -> dict[str, Any]:
    """Given an object, return a dictionary of the arguments that are
    needed by its constructor.
    """
    properties = {}
    exclude_args = ["self", "args", "kwargs"] + (exclude_fields or [])
    signature = inspect.signature(obj.__class__.__init__)
    for arg in signature.parameters:
        if arg in exclude_args:
            continue
        try:
            value = getattr(obj, arg)
        except AttributeError:
            raise AttributeError(
                f"{obj.__class__} is missing a value for {arg}, "
                f"which is needed by its constructor."
            )
        properties[arg] = value
    return properties


# pyre-fixme[24]: Generic type `type` expects 1 type parameter, use `typing.Type` to
#  avoid runtime subscripting errors.
def extract_init_args(args: dict[str, Any], class_: type) -> dict[str, Any]:
    """Given a dictionary, extract the arguments required for the
    given class's constructor.
    """
    init_args = {}
    signature = inspect.signature(class_.__init__)
    for arg, info in signature.parameters.items():
        if arg in ["self", "args", "kwargs"]:
            continue
        value = args.get(arg)
        if value is None:
            # Only necessary to raise an exception if there is no default
            # value for this argument
            if info.default is inspect.Parameter.empty:
                raise ValueError(
                    f"Cannot decode to class {class_} because required argument {arg} "
                    "is missing. If that's not the class you were intending to decode, "
                    "make sure you have updated your metric or runner registries."
                )
            else:
                # Constructor will use default value
                continue
        init_args[arg] = value
    return init_args


class SerializationMixin:
    """Base class for Ax objects that define their JSON serialization and
    deserialization logic at the class level, e.g. most commonly ``Runner``
    and ``Metric`` subclasses.

    NOTE: Using this class for Ax objects that receive other Ax objects
    as inputs, is recommended only iff the parent object (that would be
    inheriting from this base class) is not enrolled into
    CORE_ENCODER/DECODER_REGISTRY. Inheriting from this mixin with an Ax
    object that is in CORE_ENCODER/DECODER_REGISTRY, will result in a
    circular dependency, so such classes should inplement their encoding
    and decoding logic within the `json_store` module and not on the classes.

    For example, TransitionCriterion take TrialStatus as inputs and are defined
    on the CORE_ENCODER/DECODER_REGISTRY, so TransitionCriterion should not inherit
    from SerializationMixin and should define custom encoding/decoding logic within
    the json_store module.
    """

    @classmethod
    def serialize_init_args(cls, obj: SerializationMixin) -> dict[str, Any]:
        """Serialize the properties needed to initialize the object.
        Used for storage.
        """
        return serialize_init_args(obj=obj)

    @classmethod
    def deserialize_init_args(
        cls,
        args: dict[str, Any],
        decoder_registry: TDecoderRegistry | None = None,
        class_decoder_registry: TClassDecoderRegistry | None = None,
    ) -> dict[str, Any]:
        """Given a dictionary, deserialize the properties needed to initialize the
        object. Used for storage.
        """
        return extract_init_args(args=args, class_=cls)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from itertools import product

from ax.adapter.registry import Generators
from ax.analysis.plotly.sensitivity import (
    compute_sensitivity_adhoc,
    SensitivityAnalysisPlot,
)
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig
from ax.exceptions.core import UserInputError
from ax.service.ax_client import AxClient, ObjectiveProperties
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_default_generation_strategy_at_MBM_node
from pyre_extensions import assert_is_instance, none_throws


@mock_botorch_optimize
def get_test_client() -> AxClient:
    client = AxClient()
    client.create_experiment(
        is_test=True,
        name="foo",
        parameters=[
            {
                "name": "x",
                "type": "range",
                "bounds": [-1.0, 1.0],
            }
        ],
        objectives={"bar": ObjectiveProperties(minimize=True)},
    )

    for _ in range(10):
        parameterization, trial_index = client.get_next_trial()
        client.complete_trial(
            trial_index=trial_index, raw_data={"bar": parameterization["x"] ** 2}
        )
    return client


class TestSensitivityAnalysisPlot(TestCase):
    @mock_botorch_optimize
    def test_compute(self) -> None:
        client = Client()
        client.configure_experiment(
            name="foo",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        client.configure_optimization(objective="bar")

        for _ in range(6):
            for trial_index, parameterization in client.get_next_trials(
                max_trials=1
            ).items():
                client.complete_trial(
                    trial_index=trial_index,
                    raw_data={
                        "bar": assert_is_instance(parameterization["x1"], float)
                        - 2 * assert_is_instance(parameterization["x2"], float)
                    },
                )

        analysis = SensitivityAnalysisPlot(metric_names=["bar"], order="first")

        with self.assertRaisesRegex(
            UserInputError, "Must provide either a GenerationStrategy or an Adapter"
        ):
            analysis.compute()

        (card,) = analysis.compute(
            generation_strategy=client._generation_strategy
        ).flatten()
        self.assertEqual(
            card.name,
            "SensitivityAnalysisPlot",
        )
        self.assertEqual(card.title, "Sensitivity Analysis for bar")
        self.assertEqual(
            card.subtitle,
            "Understand how each parameter affects bar according to a first-order "
            "sensitivity analysis.",
        )
        self.assertEqual(
            {*card.df.columns},
            {"parameter_name", "sensitivity"},
        )
        self.assertEqual(len(card.df), 2)
        self.assertIsNotNone(card.blob)

        second_order = SensitivityAnalysisPlot(metric_names=["bar"], order="second")
        (card,) = second_order.compute(
            generation_strategy=client._generation_strategy
        ).flatten()
        self.assertEqual(len(card.df), 3)  # 2 first order + 1 second order

    @mock_botorch_optimize
    def test_compute_adhoc(self) -> None:
        metric_mapping = {"bar": "spunky"}
        client = get_test_client()
        data = client.experiment.lookup_data()
        adapter = Generators.BOTORCH_MODULAR(experiment=client.experiment, data=data)
        cards = compute_sensitivity_adhoc(
            adapter=adapter, labels=metric_mapping
        ).flatten()
        self.assertEqual(len(cards), 1)
        card = cards[0]
        self.assertEqual(card.name, "SensitivityAnalysisPlot")
        self.assertEqual(card.title, "Sensitivity Analysis for spunky")

    @mock_botorch_optimize
    @TestCase.ax_long_test(reason="Expensive to compute Sobol indicies")
    def test_online(self) -> None:
        for experiment in get_online_experiments():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            # Select an arbitrary metric from the optimization config
            metric_names = [
                none_throws(experiment.optimization_config).objective.metric_names[0]
            ]
            for order, top_k in product(["first", "second", "total"], [None, 1]):
                analysis = SensitivityAnalysisPlot(
                    metric_names=metric_names,
                    # pyre-fixme: Incompatible parameter type [6]: It isn't sure
                    # if "order" has one of the values specified by the Literal
                    order=order,
                    top_k=top_k,
                )

                _ = analysis.compute(
                    experiment=experiment, generation_strategy=generation_strategy
                )

    @mock_botorch_optimize
    @TestCase.ax_long_test(reason="Expensive to compute Sobol indicies")
    def test_offline(self) -> None:
        for experiment in get_offline_experiments():
            generation_strategy = get_default_generation_strategy_at_MBM_node(
                experiment=experiment
            )
            # Select an arbitrary metric from the optimization config
            metric_names = [
                none_throws(experiment.optimization_config).objective.metric_names[0]
            ]
            for order, top_k in product(["first", "second", "total"], [None, 1]):
                analysis = SensitivityAnalysisPlot(
                    metric_names=metric_names,
                    # pyre-fixme: Incompatible parameter type [6]: It isn't sure
                    # if "order" has one of the values specified by the Literal
                    order=order,
                    top_k=top_k,
                )

                # This prints a lot of warnings about y being constant
                # because the first MOO experiment produces constant data
                _ = analysis.compute(
                    experiment=experiment, generation_strategy=generation_strategy
                )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections import OrderedDict
from collections.abc import Iterable, Mapping
from logging import Logger

import numpy as np

import pandas as pd
import torch
from ax.adapter.adapter_utils import (
    observed_pareto_frontier as observed_pareto,
    predicted_pareto_frontier as predicted_pareto,
)
from ax.adapter.base import Adapter
from ax.adapter.cross_validation import (
    assess_model_fit,
    compute_diagnostics,
    cross_validate,
)
from ax.adapter.registry import Generators
from ax.adapter.torch import TorchAdapter
from ax.adapter.transforms.derelativize import Derelativize
from ax.core.base_trial import BaseTrial, TrialStatus
from ax.core.batch_trial import BatchTrial
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.objective import MultiObjective, Objective, ScalarizedObjective
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    OptimizationConfig,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.trial import Trial
from ax.core.types import ComparisonOp, TModelPredictArm, TParameterization
from ax.exceptions.core import UnsupportedError, UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.plot.pareto_utils import get_tensor_converter_adapter
from ax.utils.common.logger import get_logger
from botorch.utils.multi_objective.box_decompositions import DominatedPartitioning
from numpy import nan
from numpy.typing import NDArray
from pyre_extensions import assert_is_instance, none_throws

logger: Logger = get_logger(__name__)


def derelativize_opt_config(
    optimization_config: OptimizationConfig,
    experiment: Experiment,
    trial_indices: Iterable[int] | None = None,
) -> OptimizationConfig:
    tf = Derelativize(
        search_space=None, observations=None, config={"use_raw_status_quo": True}
    )
    optimization_config = tf.transform_optimization_config(
        optimization_config=optimization_config.clone(),
        adapter=get_tensor_converter_adapter(
            experiment=experiment,
            data=experiment.lookup_data(trial_indices=trial_indices),
        ),
        fixed_features=None,
    )
    return optimization_config


def get_best_raw_objective_point_with_trial_index(
    experiment: Experiment,
    optimization_config: OptimizationConfig | None = None,
    trial_indices: Iterable[int] | None = None,
) -> tuple[int, TParameterization, TModelPredictArm]:
    """Given an experiment, identifies the arm that had the best raw objective,
    based on the data fetched from the experiment.

    TModelPredictArm is of the form:
        ({metric_name: mean}, {metric_name_1: {metric_name_2: cov_1_2}})

    Note: This function will error with invalid inputs. If you would
    prefer for error logs rather than exceptions, use
    `get_best_by_raw_objective_with_trial_index`, which returns None if
    inputs are invalid.

    Args:
        experiment: Experiment, on which to identify best raw objective arm.
        optimization_config: Optimization config to use in place of the one stored
            on the experiment.
        trial_indices: Indices of trials for which to retrieve data. If None will
            retrieve data from all available trials.

    Returns:
        Tuple of trial index, parameterization, and model predictions for it.
    """
    optimization_config = optimization_config or experiment.optimization_config
    if optimization_config is None:
        raise UserInputError(
            "Cannot identify the best point without an optimization config, but no "
            "optimization config was provided on the experiment or as an argument."
        )
    if optimization_config.is_moo_problem:
        logger.warning(
            "get_best_raw_objective_point is deprecated for multi-objective "
            "optimization. This method will return an arbitrary point on the "
            "pareto frontier."
        )

    dat = experiment.lookup_data(trial_indices=trial_indices)
    if dat.df.empty:
        raise ValueError("Cannot identify best point if experiment contains no data.")
    if any(oc.relative for oc in optimization_config.all_constraints):
        optimization_config = derelativize_opt_config(
            optimization_config=optimization_config,
            experiment=experiment,
        )

    # Only COMPLETED trials should be considered when identifying the best point
    completed_indices = {
        t.index for t in experiment.trials_by_status[TrialStatus.COMPLETED]
    }
    if len(completed_indices) == 0:
        raise ValueError("Cannot identify best point if no trials are completed.")
    completed_df = dat.df[dat.df["trial_index"].isin(completed_indices)]

    is_feasible = _is_row_feasible(
        df=completed_df,
        optimization_config=optimization_config,
    )
    is_na_mask = is_feasible.isna()
    if not is_feasible.any():
        msg = (
            "No points satisfied all outcome constraints within 95 percent "
            "confidence interval."
        )
        na_arms = completed_df[is_na_mask]["arm_name"].unique()
        if len(na_arms) > 0:
            msg += (
                f" The feasibility of {len(na_arms)} arm(s) could not be determined: "
                f"{na_arms}."
            )
        raise ValueError(msg)
    # For the sake of best point identification, we only care about feasible trials.
    # The distinction between infeasible and undetermined is not important.
    is_feasible[is_na_mask] = False
    feasible_df = completed_df.loc[is_feasible]

    is_in_design = feasible_df["arm_name"].apply(
        lambda arm_name: experiment.search_space.check_membership(
            parameterization=experiment.arms_by_name[arm_name].parameters
        )
    )

    if not is_in_design.any():
        raise ValueError("No feasible points are in the search space.")

    in_design_df = feasible_df.loc[is_in_design]
    value_by_arm_pull = get_trace_by_arm_pull_from_data(
        df=in_design_df,
        optimization_config=optimization_config,
        use_cumulative_best=False,
    )

    maximize = isinstance(optimization_config.objective, MultiObjective) or (
        not optimization_config.objective.minimize
    )
    best_row_idx = (
        value_by_arm_pull["value"].idxmax()
        if maximize
        else value_by_arm_pull["value"].idxmin()
    )
    best_row = value_by_arm_pull.loc[best_row_idx]

    best_arm = experiment.arms_by_name[best_row["arm_name"]]
    best_trial_index = int(best_row["trial_index"])
    objective_rows = dat.df.loc[
        (dat.df["arm_name"] == best_arm.name)
        & (dat.df["trial_index"] == best_trial_index)
    ]
    vals = {
        row["metric_name"]: (row["mean"], row["sem"])
        for _, row in objective_rows.iterrows()
    }
    predict_arm = _raw_values_to_model_predict_arm(values=vals)

    return best_trial_index, none_throws(best_arm).parameters, predict_arm


def _extract_best_arm_from_gr(
    gr: GeneratorRun,
    trials: Mapping[int, BaseTrial],
) -> tuple[int, TParameterization, TModelPredictArm | None] | None:
    """Extracts best arm predictions from a GeneratorRun, if available,
    and maps it to the trial index of the first trial that contains it.

    Args:
        gr: GeneratorRun, from which to extract best arm predictions.
        trials: Trials from the experiment, used to map the arm to a trial index.

    Returns:
        If the best arm or the best arm predictions are not available, returns
        None. Otherwise, returns a tuple of the trial index, parameterization,
        and model predictions for the best arm.
    """
    if gr.best_arm_predictions is None:
        return None

    best_arm, best_arm_predictions = gr.best_arm_predictions

    if best_arm is None:
        return None

    for trial_index, trial in trials.items():
        if best_arm in trial.arms:
            return trial_index, best_arm.parameters, best_arm_predictions


def _raw_values_to_model_predict_arm(
    values: dict[str, tuple[float, float]],
) -> TModelPredictArm:
    return (
        {k: v[0] for k, v in values.items()},  # v[0] is mean
        {k: {k: v[1] * v[1]} for k, v in values.items()},  # v[1] is sem
    )


def get_best_parameters_from_model_predictions_with_trial_index(
    experiment: Experiment,
    adapter: Adapter | None,
    optimization_config: OptimizationConfig | None = None,
    trial_indices: Iterable[int] | None = None,
) -> tuple[int, TParameterization, TModelPredictArm | None] | None:
    """Given an experiment, returns the best predicted parameterization and
    corresponding prediction.

    The best point & predictions are computed using the given ``Adapter``
    and its ``predict`` method (if implemented). If ``adapter`` is not a
    ``TorchAdapter``, the best point is extracted from the (first) generator run
    of the latest trial. If the latest trial doesn't have a generator run, returns
    None. If the model fit assessment returns bad fit for any of the metrics, this
    will fall back to returning the best point based on raw observations.

    TModelPredictArm is of the form:
        ({metric_name: mean}, {metric_name_1: {metric_name_2: cov_1_2}})

    Args:
        experiment: ``Experiment``, on which to identify best raw objective arm.
        adapter: The ``Adapter`` to use to get the model predictions. If None, the
            best point will be extracted from the generator run of the latest trial.
        optimization_config: Optional ``OptimizationConfig`` override, to use in place
            of the one stored on the experiment.
        trial_indices: Indices of trials for which to retrieve data. If None will
            retrieve data from all available trials.

    Returns:
        Tuple of trial index, parameterization, and model predictions for it.
    """
    optimization_config = optimization_config or experiment.optimization_config
    if optimization_config is None:
        raise ValueError(
            "Cannot identify the best point without an optimization config, but no "
            "optimization config was provided on the experiment or as an argument."
        )
    if optimization_config.is_moo_problem:
        logger.warning(
            "get_best_parameters_from_model_predictions is deprecated for "
            "multi-objective optimization configs. This method will return an "
            "arbitrary point on the pareto frontier."
        )
    gr = None
    data = experiment.lookup_data(trial_indices=trial_indices)
    # Extract the latest GR from the experiment.
    for _, trial in sorted(experiment.trials.items(), key=lambda x: x[0], reverse=True):
        if isinstance(trial, Trial):
            gr = trial.generator_run
        elif isinstance(trial, BatchTrial):
            if len(trial.generator_run_structs) > 0:
                # In theory batch_trial can have >1 gr, grab the first
                gr = trial.generator_run_structs[0].generator_run
        if gr is not None:
            break

    if not isinstance(adapter, TorchAdapter):
        if gr is None:
            return None
        return _extract_best_arm_from_gr(gr=gr, trials=experiment.trials)

    # Check to see if the adapter is worth using.
    cv_results = cross_validate(model=adapter)
    diagnostics = compute_diagnostics(result=cv_results)
    assess_model_fit_results = assess_model_fit(diagnostics=diagnostics)
    objective_name = optimization_config.objective.metric.name
    # If model fit is bad use raw results
    if objective_name in assess_model_fit_results.bad_fit_metrics_to_fisher_score:
        logger.warning("Model fit is poor; falling back on raw data for best point.")

        if not _is_all_noiseless(df=data.df, metric_name=objective_name):
            logger.warning(
                "Model fit is poor and data on objective metric "
                + f"{objective_name} is noisy; interpret best points "
                + "results carefully."
            )

        return get_best_by_raw_objective_with_trial_index(
            experiment=experiment,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
        )

    res = adapter.model_best_point()
    if res is None:
        if gr is None:
            return None
        return _extract_best_arm_from_gr(gr=gr, trials=experiment.trials)

    best_arm, best_arm_predictions = res

    # Map the arm to the trial index of the first trial that contains it.
    for trial_index, trial in experiment.trials.items():
        if best_arm in trial.arms:
            return (
                trial_index,
                none_throws(best_arm).parameters,
                best_arm_predictions,
            )

    return None


def get_best_by_raw_objective_with_trial_index(
    experiment: Experiment,
    optimization_config: OptimizationConfig | None = None,
    trial_indices: Iterable[int] | None = None,
) -> tuple[int, TParameterization, TModelPredictArm] | None:
    """Given an experiment, identifies the arm that had the best raw objective,
    based on the data fetched from the experiment.

    TModelPredictArm is of the form:
        ({metric_name: mean}, {metric_name_1: {metric_name_2: cov_1_2}})

    This is a version of `get_best_raw_objective_point_with_trial_index` that
    logs errors rather than letting exceptions be raised.

    Args:
        experiment: Experiment, on which to identify best raw objective arm.
        optimization_config: Optimization config to use in place of the one stored
            on the experiment.
        trial_indices: Indices of trials for which to retrieve data. If None will
            retrieve data from all available trials.

    Returns:
        Tuple of trial index, parameterization, and model predictions for it.
    """
    try:
        result = get_best_raw_objective_point_with_trial_index(
            experiment=experiment,
            optimization_config=optimization_config,
            trial_indices=trial_indices,
        )
    except ValueError as err:
        logger.error(
            "Encountered error while trying to identify the best point: "
            f"'{err}'. Returning None."
        )
        return None
    return result


def get_pareto_optimal_parameters(
    experiment: Experiment,
    generation_strategy: GenerationStrategy,
    optimization_config: OptimizationConfig | None = None,
    trial_indices: Iterable[int] | None = None,
    use_model_predictions: bool = True,
) -> dict[int, tuple[TParameterization, TModelPredictArm]]:
    """Identifies the best parameterizations tried in the experiment so far,
    using model predictions if ``use_model_predictions`` is true and using
    observed values from the experiment otherwise. By default, uses model
    predictions to account for observation noise.

    NOTE: The format of this method's output is as follows:
    { trial_index --> (parameterization, (means, covariances) }, where means
    are a dictionary of form { metric_name --> metric_mean } and covariances
    are a nested dictionary of form
    { one_metric_name --> { another_metric_name: covariance } }.

    Args:
        experiment: Experiment, from which to find Pareto-optimal arms.
        generation_strategy: Generation strategy containing the adapter.
        optimization_config: Optimization config to use in place of the one stored
            on the experiment.
        trial_indices: Indices of trials for which to retrieve data. If None will
            retrieve data from all available trials.
        use_model_predictions: Whether to extract the Pareto frontier using
            model predictions or directly observed values. If ``True``,
            the metric means and covariances in this method's output will
            also be based on model predictions and may differ from the
            observed values.

    Returns:
        A mapping from trial index to the tuple of:
        - the parameterization of the arm in that trial,
        - two-item tuple of metric means dictionary and covariance matrix
            (model-predicted if ``use_model_predictions=True`` and observed
            otherwise).
    """
    optimization_config = optimization_config or experiment.optimization_config
    if optimization_config is None:
        raise ValueError(
            "Cannot identify the best point without an optimization config, but no "
            "optimization config was provided on the experiment or as an argument."
        )

    # Validate aspects of the experiment: that it is a MOO experiment and
    # that the current model can be used to produce the Pareto frontier.
    if not optimization_config.is_moo_problem:
        raise UnsupportedError(
            "Please use `get_best_parameters` for single-objective problems."
        )

    moo_optimization_config = assert_is_instance(
        optimization_config,
        MultiObjectiveOptimizationConfig,
    )

    # Use existing adapter if it supports MOO otherwise create a new MOO adapter
    # to use for Pareto frontier extraction.
    adapter = generation_strategy.adapter
    is_moo_adapter = (
        adapter
        and isinstance(adapter, TorchAdapter)
        and assert_is_instance(
            adapter,
            TorchAdapter,
        ).is_moo_problem
    )
    if is_moo_adapter:
        generation_strategy._curr._fit(experiment=experiment)
    else:
        adapter = Generators.BOTORCH_MODULAR(
            experiment=experiment,
            data=assert_is_instance(
                experiment.lookup_data(trial_indices=trial_indices),
                Data,
            ),
        )
    adapter = assert_is_instance(adapter, TorchAdapter)

    objective_thresholds_override = None
    # If objective thresholds are not specified in optimization config, infer them.
    if not moo_optimization_config.objective_thresholds:
        objective_thresholds_override = adapter.infer_objective_thresholds(
            search_space=experiment.search_space,
            optimization_config=optimization_config,
            fixed_features=None,
        )
        logger.info(
            f"Using inferred objective thresholds: {objective_thresholds_override}, "
            "as objective thresholds were not specified as part of the optimization "
            "configuration on the experiment."
        )

    pareto_util = predicted_pareto if use_model_predictions else observed_pareto
    pareto_optimal_observations = pareto_util(
        adapter=adapter,
        optimization_config=moo_optimization_config,
        objective_thresholds=objective_thresholds_override,
    )

    # Insert observations into OrderedDict in order of descending individual
    # hypervolume, formated as
    # { trial_index --> (parameterization, (means, covariances) }
    res: dict[int, tuple[TParameterization, TModelPredictArm]] = OrderedDict()
    for obs in pareto_optimal_observations:
        res[int(none_throws(obs.features.trial_index))] = (
            obs.features.parameters,
            (obs.data.means_dict, obs.data.covariance_matrix),
        )

    return res


def _is_row_feasible(
    df: pd.DataFrame,
    optimization_config: OptimizationConfig,
    undetermined_value: bool | None = None,
) -> pd.Series:
    """Return a series of boolean values indicating whether arms satisfy outcome
    constraints or not.

    Looks at all arm data collected and returns False for rows corresponding to arms in
    which one or more of their associated metrics' 95% confidence interval
    falls outside of any outcome constraint's bounds (i.e. we are 95% sure the
    bound is not satisfied), else True.

    Args:
        df: Dataframe of arm data, with columns "metric_name", "mean", "sem", and
            "arm_name".
        optimization_config: Optimization config to extract constraints from.
        undetermined_value: Optional value to use for rows in which the feasibility
            cannot be determined.

    Returns:
        Series of optional Boolean values indicating whether arms satisfy outcome
        constraints, or `undetermined_value` if the feasibility of an arm cannot be
        determined.
    """
    if len(optimization_config.all_constraints) < 1:
        return pd.Series([True] * len(df), index=df.index)

    relative_constraints = [
        c for c in optimization_config.all_constraints if c.relative
    ]
    if len(relative_constraints) > 0:
        logger.warning(
            f"Determining trial feasibility only supported with a derelativized "
            "OptimizationConfig, but found the following relative constraints: "
            f"{relative_constraints}. "
            f"Returning {undetermined_value} as the feasibility."
        )
        return pd.Series([undetermined_value for _ in df.index], index=df.index)

    name = df["metric_name"]

    # When SEM is NaN we should treat it as if it were 0
    sems = none_throws(df["sem"].fillna(0))
    # Bounds computed for 95% confidence interval on Normal distribution
    lower_bound = df["mean"] - sems * 1.96
    upper_bound = df["mean"] + sems * 1.96
    # TODO: Support scalarized outcome constraints by getting weights and scalarizing
    # the bounds here.

    # Nested function from OC -> Mask for consumption in later map/reduce from
    # [OC] -> Mask. Constraint relativity is handled inside so long as relative bounds
    # are set in surrounding closure (which will occur in proper experiment setup).
    def compute_feasibility_per_constraint(
        oc: OutcomeConstraint,
        lower_bound: pd.Series = lower_bound,
        upper_bound: pd.Series = upper_bound,
        name: pd.Series = name,
    ) -> pd.Series:
        name_match_mask = name == oc.metric.name
        # Return True if metrics are different, or whether the confidence
        # interval is entirely not within the bound
        if oc.op == ComparisonOp.GEQ:
            return ~name_match_mask | (upper_bound >= float(oc.bound))
        else:
            return ~name_match_mask | (lower_bound <= float(oc.bound))

    # Keep track of whether arms have mising values (NaNs) or rows.
    is_na_mask = df["mean"].isna()

    # If an arm doesn't have data for all constrained metrics, and the observed
    # constrained metric values are feasible, (in)feasibility cannot be determined
    # conclusively.
    has_missing_metric_mask = pd.Series([False] * len(df), index=df.index)
    constrained_metric_names = {
        oc.metric.name for oc in optimization_config.all_constraints
    }
    for arm_name, arm_group in df.groupby("arm_name"):
        metrics_for_arm = set(arm_group["metric_name"].unique())
        missing_metrics = constrained_metric_names - metrics_for_arm
        if missing_metrics:
            logger.warning(
                f"Arm {arm_name} is missing data for one or more constrained metrics: "
                f"{missing_metrics}."
            )
            has_missing_metric_mask = has_missing_metric_mask | (
                df["arm_name"] == arm_name
            )

    # Computing feasibility on a per-row (metric-arm-combination) basis.
    is_feasible_per_constraint_list = [
        compute_feasibility_per_constraint(oc=oc)
        for oc in optimization_config.all_constraints
    ]
    # stacking the feasibility masks for each constraint an checking if all are feasible
    is_feasible_mask = pd.DataFrame(is_feasible_per_constraint_list).all(axis=0)

    # can definititively determine infeasibility for all rows that are evaluated
    # infeasible (~is_feasible_mask) based on available data (~is_na_mask).
    infeasible_df = df[~is_feasible_mask & ~is_na_mask]
    infeasible_arm_names = set(infeasible_df["arm_name"].unique())
    # Can't determine feasibility for rows that are not definitively infeasible
    # and that have missing values (NaN or missing metrics).
    na_df = df[has_missing_metric_mask | is_na_mask]
    na_arm_names = set(na_df["arm_name"].unique())

    def tag_feasible_arms(
        x: str,
        infeasible_arm_names: set[str] = infeasible_arm_names,
        na_arm_names: set[str] = na_arm_names,
    ) -> bool | None:
        if x in infeasible_arm_names:
            return False
        elif x in na_arm_names:
            return undetermined_value
        return True

    return assert_is_instance(
        df["arm_name"].apply(tag_feasible_arms),
        pd.Series,
    )


def _is_all_noiseless(df: pd.DataFrame, metric_name: str) -> bool:
    """Noiseless is defined as SEM = 0 or SEM = NaN on a given metric (usually
    the objective).
    """

    name_mask = df["metric_name"] == metric_name
    df_metric_arms_sems = df[name_mask]["sem"]

    return ((df_metric_arms_sems == 0) | df_metric_arms_sems == nan).all()


def get_values_of_outcomes_single_or_scalarized_objective(
    df_wide: pd.DataFrame, objective: Objective
) -> NDArray:
    """
    Return a list with one entry for each row in `df_wide` according to the
    objective `objective` and whether the outcomes are feasible.

    Whether higher or lower is better depends on `objective.minimize` (no
    absolute values are taken here).

    The entry for any infeasible value will be infinity if the objective is to
    minimize and negative infinity if the objective is to maximize.

    Example:
    >>> objective = Objective(metric=Metric(name="m1"), minimize=True)
    >>> df_wide = pd.DataFrame.from_records(
    ...     [
    ...         {"m1": 2.0, "feasible": True},
    ...         {"m1": 1.0, "feasible": False},
    ...     ]
    ... )
    >>> get_value_of_outcomes_single_or_scalarized_objective(
    ...     df_wide=df_wide, objective=objective
    ... )
    np.array([2.0, inf])
    """
    if isinstance(objective, MultiObjective):
        raise ValueError(
            "MultiObjective is not supported. Use "
            "`get_hypervolume_of_outcomes_multi_objective`."
        )
    if isinstance(objective, ScalarizedObjective):
        value = df_wide[objective.metric_names].dot(objective.weights).to_numpy()
    else:
        value = df_wide[objective.metric.name].to_numpy()
    value = value.astype(np.float64)
    infeasible_idx = np.where(~df_wide["feasible"])[0]
    value[infeasible_idx] = float("inf") if objective.minimize else float("-inf")
    return value


def _compute_hv_trace(
    ref_point: torch.Tensor,
    metrics_tensor: torch.Tensor,
    is_feasible_array: NDArray,
    use_cumulative_hv: bool,
) -> list[float]:
    # Compute hypervolume of feasible points
    hvs = []
    ref_point = ref_point

    if use_cumulative_hv:
        partitioning = DominatedPartitioning(ref_point=ref_point)
        cumulative_hv = 0.0
        for i, is_feasible in enumerate(is_feasible_array):
            if not is_feasible:
                hvs.append(cumulative_hv)
            else:
                Y = metrics_tensor[[i], :]
                partitioning.update(Y=Y)
                cumulative_hv = partitioning.compute_hypervolume().item()
                hvs.append(cumulative_hv)
        return hvs

    for i, is_feasible in enumerate(is_feasible_array):
        if not is_feasible:
            hvs.append(0.0)
        else:
            Y = metrics_tensor[[i], :]
            partitioning = DominatedPartitioning(ref_point=ref_point, Y=Y)
            hvs.append(partitioning.compute_hypervolume().item())
    return hvs


def get_hypervolume_trace_of_outcomes_multi_objective(
    df_wide: pd.DataFrame,
    optimization_config: MultiObjectiveOptimizationConfig,
    use_cumulative_hv: bool = True,
) -> list[float]:
    """
    Get hypervolume of the outcomes represented in `df_wide`.

    Args:
        df_wide: Dataframe with columns ["feasible"] + relevant
            metrics. This can come from reshaping the data that comes from `Data.df`.
        optimization_config: A multi-objective optimization config with a
            `MultiObjective` (not a `ScalarizedObjective`).
        use_cumulative_hv: If True, the hypervolume returned is the cumulative
            hypervolume of the points in each row. Otherwise, this is the
            hypervolume of each point.

    Returns:
        A list of hypervolumes, one for each row in `df_wide`.

    Example:
    >>> optimization_config = MultiObjectiveOptimizationConfig(
    ...     objective=MultiObjective(
    ...         objectives=[
    ...             Objective(metric=Metric(name="m1"), minimize=False),
    ...             Objective(metric=Metric(name="m2"), minimize=False),
    ...         ]
    ...     ),
    ... )
    >>> # Objective threshols will be inferred to be zero
    >>> df_wide = pd.DataFrame.from_records(
    ...     [
    ...         {"m1": 0.0, "m2": 0.0, "feasible": True},
    ...         {"m1": 1.0, "m2": 2.0, "feasible": True},
    ...         {"m1": 2.0, "m2": 1.0, "feasible": False},
    ...         {"m1": 3.0, "m2": 3.0, "feasible": True},
    ...     ]
    ... )
    >>> get_hypervolume_trace_of_outcomes_multi_objective(
    ...     df_wide=df_wide,
    ...     optimization_config=optimization_config,
    ...     use_cumulative_hv=True,
    ... )
    [0.0, 2.0, 2.0, 9.0]
    >>>
    >>> get_hypervolume_trace_of_outcomes_multi_objective(
    ...     df_wide=df_wide,
    ...     optimization_config=optimization_config,
    ...     use_cumulative_hv=False,
    ... )
    [0.0, 2.0, 0.0, 9.0]
    """
    objective = assert_is_instance(optimization_config.objective, MultiObjective)
    for obj in objective.objectives:
        if obj.minimize:
            df_wide[obj.metric.name] *= -1

    objective_thresholds = []
    objective_thresholds_dict = {
        threshold.metric.name: threshold
        for threshold in optimization_config.objective_thresholds
    }
    for obj in objective.objectives:
        metric_name = obj.metric.name
        if metric_name in objective_thresholds_dict:
            threshold = objective_thresholds_dict[metric_name]
            if threshold.relative:
                raise ValueError(
                    "Relative objective thresholds are not supported. Please "
                    "`Derelativize` the optimization config, or use "
                    "`get_trace`."
                )
            bound = threshold.bound
        else:
            metric_vals = df_wide[metric_name]
            bound = metric_vals.max() if obj.minimize else metric_vals.min()

        objective_thresholds.append(-bound if obj.minimize else bound)

    objective_thresholds = torch.tensor(objective_thresholds, dtype=torch.double)

    metrics_tensor = torch.from_numpy(df_wide[objective.metric_names].to_numpy())
    return _compute_hv_trace(
        ref_point=objective_thresholds,
        metrics_tensor=metrics_tensor,
        is_feasible_array=df_wide["feasible"].to_numpy(),
        use_cumulative_hv=use_cumulative_hv,
    )


def get_trace_by_arm_pull_from_data(
    df: pd.DataFrame,
    optimization_config: OptimizationConfig,
    use_cumulative_best: bool = True,
) -> pd.DataFrame:
    """
    Get a trace of the objective value or hypervolume of outcomes.

    An "arm pull" is the combination of a trial (index) and an arm. This
    function returns a single value for each arm pull, even if there are
    multiple arms per trial or if an arm is repeated in multiple trials.

    Args:
        df: Data in the format returned by ``Data.df``, with a separate row for
            each trial index-arm name-metric.
        optimization_config: ``OptimizationConfig`` to use to get the trace. Must
            not be in relative form.
        use_cumulative_best: If True, the trace will be the cumulative best
            objective. Otherwise, the trace will be the value of each point.

    Return:
        A DataFrame containing columns 'trial_index', 'arm_name', and "value",
        where "value" is the value of the outcomes attained.
    """
    if any(oc.relative for oc in optimization_config.all_constraints):
        raise ValueError(
            "Relativized optimization config not supported. Please "
            "`Derelativize` the optimization config, or use `get_trace`."
        )

    empty_result = pd.DataFrame(columns=["trial_index", "arm_name", "value"])

    if len(df) == 0:
        return empty_result

    # reshape data to wide, using only the metrics in the optimization config
    metrics = list(optimization_config.metrics.keys())

    df["row_feasible"] = _is_row_feasible(
        df=df,
        optimization_config=optimization_config,
        # For the sake of this function, we only care about feasible trials. The
        # distinction between infeasible and undetermined is not important.
        undetermined_value=False,
    )

    # Transform to a DataFrame with columns ["trial_index", "arm_name"] +
    # relevant metric names, and values being means.
    df_wide = (
        df[df["metric_name"].isin(metrics)]
        .set_index(["trial_index", "arm_name", "metric_name"])["mean"]
        .unstack(level="metric_name")
    )
    missing_metrics = [
        m for m in metrics if m not in df_wide.columns or df_wide[m].isnull().any()
    ]
    if len(missing_metrics) > 0:
        raise ValueError(
            "Some metrics are not present for all trials and arms. The "
            f"following are missing: {missing_metrics}."
        )
    if len(df_wide) == 0:
        return empty_result
    df_wide["feasible"] = df.groupby(["trial_index", "arm_name"])["row_feasible"].all()
    df_wide.reset_index(inplace=True)

    # MOO and *not* ScalarizedObjective
    if isinstance(optimization_config.objective, MultiObjective):
        optimization_config = assert_is_instance(
            optimization_config, MultiObjectiveOptimizationConfig
        )
        df_wide["value"] = get_hypervolume_trace_of_outcomes_multi_objective(
            df_wide=df_wide,
            optimization_config=optimization_config,
            use_cumulative_hv=use_cumulative_best,
        )
        return df_wide[["trial_index", "arm_name", "value"]]
    df_wide["value"] = get_values_of_outcomes_single_or_scalarized_objective(
        df_wide=df_wide, objective=optimization_config.objective
    )
    if df_wide["feasible"].any() and use_cumulative_best:
        min_or_max = (
            np.minimum if optimization_config.objective.minimize else np.maximum
        )
        df_wide["value"] = min_or_max.accumulate(df_wide["value"])
    return df_wide[["trial_index", "arm_name", "value"]]


def get_trace(
    experiment: Experiment,
    optimization_config: OptimizationConfig | None = None,
) -> list[float]:
    """Compute the optimization trace at each iteration.

    Given an experiment and an optimization config, compute the performance
    at each iteration. For multi-objective, the performance is computed as
    the hypervolume. For single objective, the performance is computed as
    the best observed objective value.

    Infeasible points (that violate constraints) do not contribute to
    improvements in the optimization trace. If the first trial(s) are infeasible,
    the trace can start at inf or -inf.

    An iteration here refers to a completed or early-stopped (batch) trial.
    There will be one performance metric in the trace for each iteration.

    Args:
        experiment: The experiment to get the trace for.
        optimization_config: Optimization config to use in place of the one
            stored on the experiment.

    Returns:
        A list of performance values at each iteration.
    """
    optimization_config = optimization_config or none_throws(
        experiment.optimization_config
    )
    df = experiment.lookup_data().df
    if len(df) == 0:
        return []
    # Get the names of the metrics in optimization config.
    metric_names = set(optimization_config.objective.metric_names)
    for cons in optimization_config.outcome_constraints:
        metric_names.update({cons.metric.name})
    metric_names = list(metric_names)

    # Don't compute results for status quo data (for compatibility with legacy behavior)
    trial_is_completed = df["trial_index"].map(
        {
            i: t.status in (TrialStatus.COMPLETED, TrialStatus.EARLY_STOPPED)
            for i, t in experiment.trials.items()
        }
    )
    idx = df["metric_name"].isin(metric_names) & trial_is_completed
    # Don't include status quo (for compatibility with legacy behavior)
    if (status_quo := experiment.status_quo) is not None:
        idx &= df["arm_name"] != status_quo.name
    df = df.loc[idx, :]
    if len(df) == 0:
        return []

    # Derelativize the optimization config.
    optimization_config = derelativize_opt_config(
        optimization_config=optimization_config,
        experiment=experiment,
    )

    # Get a value for each trial_index + arm
    value_by_arm_pull = get_trace_by_arm_pull_from_data(
        df=df,
        optimization_config=optimization_config,
        use_cumulative_best=True,
    )
    # Aggregate to trial level
    objective = optimization_config.objective
    maximize = isinstance(objective, MultiObjective) or not objective.minimize
    trial_grouped = value_by_arm_pull.groupby("trial_index")["value"]
    if maximize:
        value_by_trial = trial_grouped.max()
        cumulative_value = np.maximum.accumulate(value_by_trial)
    else:
        value_by_trial = trial_grouped.min()
        cumulative_value = np.minimum.accumulate(value_by_trial)

    return cumulative_value.tolist()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import numpy as np
from ax.benchmark.benchmark_result import BenchmarkResult
from ax.utils.common.testutils import TestCase

from ax.utils.testing.core_stubs import get_experiment


class TestBenchmarkResult(TestCase):
    def test_benchmark_result_invalid_inputs(self) -> None:
        """
        Test that a BenchmarkResult cannot be specified with both an `experiment`
        and an `experiment_storage_id`.
        """
        with self.assertRaisesRegex(ValueError, "Cannot specify both an `experiment` "):
            BenchmarkResult(
                name="name",
                seed=0,
                inference_trace=np.array([]),
                oracle_trace=np.array([]),
                optimization_trace=np.array([]),
                score_trace=np.array([]),
                cost_trace=np.array([]),
                fit_time=0.0,
                gen_time=0.0,
                experiment=get_experiment(),
                experiment_storage_id="experiment_storage_id",
            )

        with self.assertRaisesRegex(
            ValueError, "Must provide an `experiment` or `experiment_storage_id`"
        ):
            BenchmarkResult(
                name="name",
                seed=0,
                inference_trace=np.array([]),
                oracle_trace=np.array([]),
                optimization_trace=np.array([]),
                score_trace=np.array([]),
                cost_trace=np.array([]),
                fit_time=0.0,
                gen_time=0.0,
            )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import pandas as pd
from ax.analysis.search_space_summary import SearchSpaceSummary
from ax.api.client import Client
from ax.api.configs import ChoiceParameterConfig, RangeParameterConfig
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments


class TestSearchSpaceSummary(TestCase):
    def test_compute(self) -> None:
        client = Client()
        client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0.1, 1),
                    scaling="log",
                ),
                ChoiceParameterConfig(
                    name="x2",
                    parameter_type="int",
                    values=[0, 1],
                ),
            ],
        )

        analysis = SearchSpaceSummary()

        with self.assertRaisesRegex(UserInputError, "requires an `Experiment`"):
            analysis.compute()

        experiment = client._experiment
        card = analysis.compute(experiment=experiment)

        # Test metadata
        self.assertEqual(card.name, "SearchSpaceSummary")
        self.assertEqual(card.title, "SearchSpaceSummary for `test_experiment`")
        self.assertEqual(
            card.subtitle,
            (
                "The search space summary provides an overview of all "
                "parameters, including their names, types, and ranges or "
                "categories. This holistic view provides quick understanding "
                "on the parameters being optimized, allowing one to verify "
                "and adjust the search space configuration for effective "
                "exploration."
            ),
        )
        self.assertIsNotNone(card.blob)

        # Test dataframe for accuracy
        self.assertEqual(
            {*card.df.columns},
            {"Name", "Type", "Flags", "Datatype", "Domain"},
        )
        expected = pd.DataFrame(
            {
                "Name": ["x1", "x2"],
                "Type": ["Range", "Choice"],
                "Domain": ["range=[0.1, 1.0]", "values=[0, 1]"],
                "Datatype": ["float", "int"],
                "Flags": ["log_scale", "ordered, sorted"],
            }
        )
        pd.testing.assert_frame_equal(card.df, expected)

    def test_online(self) -> None:
        # Test SearchSpaceSummary can be computed for a variety of experiments which
        # resemble those we see in an online setting.

        analysis = SearchSpaceSummary()
        for experiment in get_online_experiments():
            _ = analysis.compute(experiment=experiment)

    def test_offline(self) -> None:
        # Test SearchSpaceSummary can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        analysis = SearchSpaceSummary()
        for experiment in get_offline_experiments():
            _ = analysis.compute(experiment=experiment)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-unsafe

from ax.api.configs import StorageConfig

# Allow Ax to be used without SQLAlchemy.
try:
    from ax.storage.sqa_store.decoder import Decoder
    from ax.storage.sqa_store.encoder import Encoder
    from ax.storage.sqa_store.sqa_config import SQAConfig
    from ax.storage.sqa_store.structs import DBSettings
except (ModuleNotFoundError, TypeError):
    Decoder = None
    Encoder = None
    SQAConfig = None
    DBSettings = None


def db_settings_from_storage_config(
    storage_config: StorageConfig,
) -> DBSettings:
    """Construct DBSettings (expected by WithDBSettingsBase) from StorageConfig."""
    if (bundle := storage_config.registry_bundle) is not None:
        encoder = bundle.encoder
        decoder = bundle.decoder
    else:
        encoder = Encoder(config=SQAConfig())
        decoder = Decoder(config=SQAConfig())

    return DBSettings(
        creator=storage_config.creator,
        url=storage_config.url,
        encoder=encoder,
        decoder=decoder,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

import dataclasses

from unittest.mock import patch

import numpy as np

import torch
from ax.core.search_space import RobustSearchSpaceDigest, SearchSpaceDigest
from ax.generators.torch.botorch_modular.input_constructors.input_transforms import (
    input_transform_argparse,
)
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from botorch.models.transforms.input import (
    FilterFeatures,
    InputPerturbation,
    InputStandardize,
    InputTransform,
    Normalize,
    Warp,
)
from botorch.utils.datasets import MultiTaskDataset, SupervisedDataset


class DummyInputTransform(InputTransform):  # pyre-ignore [13]
    pass


class InputTransformArgparseTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        X = torch.randn((10, 4))
        Y = torch.randn((10, 2))
        self.dataset = SupervisedDataset(
            X=X,
            Y=Y,
            feature_names=[f"x{i}" for i in range(4)],
            outcome_names=[f"y{i}" for i in range(2)],
        )
        self.search_space_digest = SearchSpaceDigest(
            feature_names=["a", "b", "c"],
            bounds=[(0.0, 1.0), (0, 2), (0, 4)],
            ordinal_features=[0],
            categorical_features=[1],
            discrete_choices={0: [0, 1, 2], 1: [0, 0.25, 4.0]},
            task_features=[2],
            fidelity_features=[0],
            target_values={0: 1.0},
            robust_digest=None,
        )

    def test_notImplemented(self) -> None:
        with self.assertRaises(NotImplementedError) as e:
            input_transform_argparse[
                type(None)
            ]  # passing `None` produces a different error
            self.assertTrue("Could not find signature for" in str(e))

    def test_register(self) -> None:
        @input_transform_argparse.register(DummyInputTransform)
        def _argparse(input_transform: DummyInputTransform) -> None:
            pass

        self.assertEqual(input_transform_argparse[DummyInputTransform], _argparse)

    def test_fallback(self) -> None:
        with patch.dict(input_transform_argparse.funcs, {}):

            @input_transform_argparse.register(InputTransform)
            def _argparse(input_transform_class: InputTransform) -> None:
                pass

            self.assertEqual(input_transform_argparse[InputTransform], _argparse)

    def test_argparse_input_transform(self) -> None:
        input_transform_kwargs = input_transform_argparse(
            InputTransform,
            dataset=self.dataset,
        )

        self.assertEqual(input_transform_kwargs, {})

        input_transform_kwargs = input_transform_argparse(
            InputStandardize, dataset=self.dataset, input_transform_options={"d": 10}
        )

        self.assertEqual(input_transform_kwargs, {"d": 10})

    def test_argparse_normalize(self) -> None:
        input_transform_kwargs = input_transform_argparse(
            Normalize,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
        )

        self.assertTrue(
            torch.all(
                torch.isclose(
                    input_transform_kwargs["bounds"],
                    torch.tensor(
                        [[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]],
                        dtype=input_transform_kwargs["bounds"].dtype,
                    ),
                )
            )
        )
        # Use the dimension of the datasets as the ground truth as
        # that's what will be fed to the model.
        self.assertEqual(input_transform_kwargs["d"], 4)
        self.assertEqual(input_transform_kwargs["indices"], [0, 1, 3])

        input_transform_kwargs = input_transform_argparse(
            Normalize,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={
                "d": 4,
                "bounds": torch.tensor(
                    [[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]], dtype=torch.float64
                ),
            },
        )

        self.assertEqual(input_transform_kwargs["d"], 4)
        self.assertEqual(input_transform_kwargs["indices"], [0, 1, 3])

        self.assertTrue(
            torch.all(
                torch.isclose(
                    input_transform_kwargs["bounds"],
                    torch.tensor(
                        [[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]], dtype=torch.float64
                    ),
                )
            )
        )

        # Test with MultiTaskDataset.
        dataset1 = SupervisedDataset(
            X=torch.cat([torch.rand(5, 4), torch.zeros(5, 1)], dim=-1),
            Y=torch.randn(5, 1),
            feature_names=[f"x{i}" for i in range(4)] + [Keys.TASK_FEATURE_NAME.value],
            outcome_names=["y0"],
        )

        dataset2 = SupervisedDataset(
            X=torch.cat([torch.rand(5, 2), torch.zeros(5, 1)], dim=-1),
            Y=torch.randn(5, 1),
            feature_names=[f"x{i}" for i in range(2)] + [Keys.TASK_FEATURE_NAME.value],
            outcome_names=["y1"],
        )
        mtds = MultiTaskDataset(
            datasets=[dataset1, dataset2],
            target_outcome_name="y0",
            task_feature_index=-1,
        )
        mt_ssd = dataclasses.replace(
            self.search_space_digest,
            feature_names=self.search_space_digest.feature_names
            + [Keys.TASK_FEATURE_NAME.value],
            task_features=[-1],
            bounds=self.search_space_digest.bounds + [(0.0, 1.0)],
        )
        input_transform_kwargs = input_transform_argparse(
            Normalize,
            dataset=mtds,
            search_space_digest=mt_ssd,
        )
        self.assertEqual(input_transform_kwargs["d"], 5)
        # task feature should be omitted
        self.assertEqual(input_transform_kwargs["indices"], [0, 1, 2, 3])

        input_transform_kwargs = input_transform_argparse(
            Normalize,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={
                "bounds": None,
            },
        )

        self.assertEqual(input_transform_kwargs["d"], 4)
        self.assertEqual(input_transform_kwargs["indices"], [0, 1, 3])
        self.assertTrue(input_transform_kwargs["bounds"] is None)

    def test_argparse_warp(self) -> None:
        self.search_space_digest.task_features = [0, 3]
        input_transform_kwargs = input_transform_argparse(
            Warp,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
        )

        self.assertEqual(input_transform_kwargs["indices"], [1, 2])
        self.assertEqual(input_transform_kwargs["d"], 4)
        self.assertTrue(
            torch.equal(
                input_transform_kwargs["bounds"],
                torch.tensor([[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]]),
            )
        )

        input_transform_kwargs = input_transform_argparse(
            Warp,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={"indices": [0, 1]},
        )
        self.assertEqual(
            input_transform_kwargs["indices"],
            [0, 1],
        )
        self.assertEqual(input_transform_kwargs["d"], 4)
        self.assertTrue(
            torch.equal(
                input_transform_kwargs["bounds"],
                torch.tensor([[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]]),
            )
        )
        input_transform_kwargs = input_transform_argparse(
            Warp,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={"indices": [0, 1]},
            torch_dtype=torch.float64,
        )
        self.assertTrue(
            torch.equal(
                input_transform_kwargs["bounds"],
                torch.tensor([[0.0, 0.0, 0.0], [1.0, 2.0, 4.0]], dtype=torch.float64),
            )
        )

    def test_argparse_input_perturbation(self) -> None:
        self.search_space_digest.robust_digest = RobustSearchSpaceDigest(
            sample_param_perturbations=lambda: np.zeros((2, 2)),
        )

        input_transform_kwargs = input_transform_argparse(
            InputPerturbation,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
        )

        self.assertEqual(input_transform_kwargs["multiplicative"], False)

        self.assertTrue(
            torch.all(
                torch.isclose(
                    input_transform_kwargs["perturbation_set"],
                    torch.tensor([[0.0, 0.0], [0.0, 0.0]], dtype=torch.float64),
                )
            )
        )

    def test_argparse_filter_features(self) -> None:
        # Test with no input_transform_options - should return all feature indices
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
        )
        all_feature_indices = torch.arange(
            len(self.dataset.feature_names), dtype=torch.int64
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(input_transform_kwargs["feature_indices"], all_feature_indices)
        )

        # Test with empty input_transform_options - should return all feature indices
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={},
        )
        all_feature_indices = torch.arange(
            len(self.dataset.feature_names), dtype=torch.int64
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(input_transform_kwargs["feature_indices"], all_feature_indices)
        )

        # Test with explicit feature_indices - should pass through unchanged
        feature_indices = torch.tensor([0, 2, 3], dtype=torch.int64)
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={"feature_indices": feature_indices},
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(input_transform_kwargs["feature_indices"], feature_indices)
        )
        # Test with ignored_params - should convert to feature_indices
        ignored_params = ["x1", "x3"]
        expected_feature_indices = torch.tensor(
            [0, 2], dtype=torch.int64
        )  # Keep x0 and x2, ignore x1 and x3
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={"ignored_params": ignored_params},
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(
                input_transform_kwargs["feature_indices"], expected_feature_indices
            )
        )

        # Test with empty ignored_params - should return all feature indices
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={"ignored_params": []},
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(input_transform_kwargs["feature_indices"], all_feature_indices)
        )

        # Test when both feature_indices and ignored_params are specified - and
        # they are consistent
        feature_indices = torch.tensor([0, 2, 3], dtype=torch.int64)
        input_transform_kwargs = input_transform_argparse(
            FilterFeatures,
            dataset=self.dataset,
            search_space_digest=self.search_space_digest,
            input_transform_options={
                "feature_indices": feature_indices,
                "ignored_params": ["x1"],
            },
        )
        self.assertEqual(set(input_transform_kwargs.keys()), {"feature_indices"})
        self.assertTrue(
            torch.equal(input_transform_kwargs["feature_indices"], feature_indices)
        )

        # Test when both feature_indices and ignored_params are specified - and
        # they are inconsistent
        feature_indices = torch.tensor([0, 2, 3], dtype=torch.int64)
        with self.assertRaisesRegex(
            ValueError,
            r"Filtered features passed in by feature_indices .* "
            r"is inconsistent with filtered feature indices computed from "
            r"ignored_params",
        ):
            input_transform_kwargs = input_transform_argparse(
                FilterFeatures,
                dataset=self.dataset,
                search_space_digest=self.search_space_digest,
                input_transform_options={
                    "feature_indices": feature_indices,
                    "ignored_params": ["x0", "x1"],
                },
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import enum
from collections.abc import Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass

from logging import Logger
from typing import Any, Union

from ax.core.arm import Arm
from ax.core.auxiliary import AuxiliaryExperiment, AuxiliaryExperimentPurpose
from ax.core.experiment import DataType, Experiment
from ax.core.metric import Metric
from ax.core.multi_type_experiment import MultiTypeExperiment
from ax.core.objective import MultiObjective, Objective
from ax.core.observation import ObservationFeatures
from ax.core.optimization_config import (
    MultiObjectiveOptimizationConfig,
    ObjectiveThreshold,
    OptimizationConfig,
)
from ax.core.outcome_constraint import OutcomeConstraint
from ax.core.parameter import (
    ChoiceParameter,
    DerivedParameter,
    FixedParameter,
    Parameter,
    PARAMETER_PYTHON_TYPE_MAP,
    ParameterType,
    RangeParameter,
    TParameterType,
)
from ax.core.parameter_constraint import (
    OrderConstraint,
    ParameterConstraint,
    validate_constraint_parameters,
)
from ax.core.runner import Runner
from ax.core.search_space import HierarchicalSearchSpace, SearchSpace
from ax.core.types import ComparisonOp, TParameterization, TParamValue
from ax.exceptions.core import UnsupportedError
from ax.utils.common.constants import Keys
from ax.utils.common.logger import get_logger
from ax.utils.common.typeutils import (
    assert_is_instance_of_tuple,
    assert_is_instance_optional,
)
from pyre_extensions import assert_is_instance, none_throws


logger: Logger = get_logger(__name__)


"""Utilities for RESTful-like instantiation of Ax classes needed in AxClient."""


TParameterRepresentation = dict[
    str, Union[TParamValue, Sequence[TParamValue], dict[str, list[str]]]
]
PARAM_CLASSES = ["range", "choice", "fixed", "derived"]
PARAM_TYPES = {"int": int, "float": float, "bool": bool, "str": str}
COMPARISON_OPS: dict[str, ComparisonOp] = {
    "<=": ComparisonOp.LEQ,
    ">=": ComparisonOp.GEQ,
}
EXPECTED_KEYS_IN_PARAM_REPR = {
    "name",
    "type",
    "values",
    "bounds",
    "value",
    "value_type",
    "log_scale",
    "target_value",
    "is_fidelity",
    "sort_values",
    "is_ordered",
    "is_task",
    "digits",
    "dependents",
    "expression_str",
}


INVALID_CONSTRAINT_ERROR_MSG = (
    "Received invalid parameter constraint format: `{}`. "
    "Please use one of the following forms:\n"
    "* Ordered constraints: `<p1> >= <p2>` or `<p1> <= <p2>`, where `<p1>` and `<p2>` "
    "are parameter names.\n"
    "* Unweighted linear constraints: `<p1> >= <b>` or `<p1> + <p2> <= <b>`, where you "
    "can add one or more parameters on the left side, and `<b>` is a numeric value.\n"
    "* Weighted linear constraints: `<w1>*<p1> >= <b>` or "
    "`<w1>*<p1> + <w2>*<p2> <= <b>`, where you can add one or more weighted terms on "
    "the left side, and there should be no spaces between weights and parameter "
    'names.\nAcceptable comparison operators are ">=" and "<=".'
)


class MetricObjective(enum.Enum):
    MINIMIZE = enum.auto()
    MAXIMIZE = enum.auto()


@dataclass
class ObjectiveProperties:
    r"""Class that holds properties of objective functions. Can be used to define an
    the `objectives` argument of ax_client.create_experiment, e.g.:

        ax_client.create_experiment(
            name="moo_experiment",
            parameters=[...],
            objectives={
                # `threshold` arguments are optional
                "a": ObjectiveProperties(minimize=False, threshold=ref_point[0]),
                "b": ObjectiveProperties(minimize=False, threshold=ref_point[1]),
            },
        )

    Args:
        - minimize: Boolean indicating whether the objective is to be minimized
            or maximized.
        - threshold: Optional `float` representing the smallest objective value
            (resp. largest if minimize=True) that is considered valuable in the context
            of multi-objective optimization. In BoTorch and in the literature, this is
            also known as an element of the reference point vector that defines the
            hyper-volume of the Pareto front.
    """

    minimize: bool
    threshold: float | None = None


@dataclass(frozen=True)
class FixedFeatures:
    """Class for representing fixed features via the Service API."""

    parameters: TParameterization
    trial_index: int | None = None


class InstantiationBase:
    """
    This is a lightweight stateless class that bundles together instantiation utils.
    It is used both on its own and as a mixin to AxClient, with the intent that
    these methods can be overridden by its subclasses for specific use cases.
    """

    @staticmethod
    def _get_deserialized_metric_kwargs(
        metric_class: type[Metric],
        name: str,
        metric_definitions: dict[str, dict[str, Any]] | None,
    ) -> tuple[type[Metric], dict[str, Any]]:
        """Get metric kwargs from metric_definitions if available and deserialize
        if so.  Deserialization is necessary because they were serialized on creation"""
        # deepcopy is used because of subsequent modifications to the dict
        metric_kwargs = deepcopy((metric_definitions or {}).get(name, {}))
        metric_class = metric_kwargs.pop("metric_class", metric_class)
        # this is necessary before deserialization because name will be required
        metric_kwargs["name"] = metric_kwargs.get("name", name)
        metric_kwargs = metric_class.deserialize_init_args(metric_kwargs)
        return metric_class, metric_kwargs

    @classmethod
    def _make_metric(
        cls,
        name: str,
        lower_is_better: bool | None = None,
        metric_class: type[Metric] = Metric,
        for_opt_config: bool = False,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> Metric:
        if " " in name:
            raise ValueError(
                "Metric names cannot contain spaces when used with AxClient. Got "
                f"{name!r}."
            )

        metric_definitions = metric_definitions or {}

        metric_class, kwargs = cls._get_deserialized_metric_kwargs(
            name=name,
            metric_definitions=metric_definitions,
            metric_class=metric_class,
        )
        # avoid conflict is lower_is_better is specified in kwargs
        kwargs["lower_is_better"] = kwargs.get("lower_is_better", lower_is_better)
        return metric_class(
            **kwargs,
        )

    @staticmethod
    def _get_parameter_type(python_type: TParameterType) -> ParameterType:
        for param_type, py_type in PARAMETER_PYTHON_TYPE_MAP.items():
            if py_type is python_type:
                return param_type
        raise ValueError(f"No AE parameter type corresponding to {python_type}.")

    @classmethod
    def _to_parameter_type(
        cls,
        vals: list[TParamValue],
        typ: str | None,
        param_name: str,
        field_name: str,
    ) -> ParameterType:
        if typ is None:
            typ = type(none_throws(vals[0]))
            parameter_type = cls._get_parameter_type(typ)  # pyre-ignore[6]
            assert all(isinstance(x, typ) for x in vals), (
                f"Values in `{field_name}` not of the same type and no "
                "`value_type` was explicitly specified; cannot infer "
                f"value type for parameter {param_name}."
            )
            logger.info(
                f"Inferred value type of {parameter_type} for parameter {param_name}. "
                "If that is not the expected value type, you can explicitly specify "
                "'value_type' ('int', 'float', 'bool' or 'str') in parameter dict."
            )
            return parameter_type
        return cls._get_parameter_type(PARAM_TYPES[typ])  # pyre-ignore[6]

    @classmethod
    def _make_range_param(
        cls,
        name: str,
        representation: TParameterRepresentation,
        parameter_type: str | None,
    ) -> RangeParameter:
        assert "bounds" in representation, "Bounds are required for range parameters."
        bounds = representation["bounds"]
        assert isinstance(bounds, list) and len(bounds) == 2, (
            f"Cannot parse parameter {name}: for range parameters, json representation "
            "should include a list of two values, lower and upper bounds of the range."
        )
        return RangeParameter(
            name=name,
            parameter_type=cls._to_parameter_type(
                bounds, parameter_type, name, "bounds"
            ),
            lower=assert_is_instance_of_tuple(bounds[0], (float, int)),
            upper=assert_is_instance_of_tuple(bounds[1], (float, int)),
            log_scale=assert_is_instance(representation.get("log_scale", False), bool),
            digits=representation.get("digits", None),  # pyre-ignore[6]
            is_fidelity=assert_is_instance(
                representation.get("is_fidelity", False), bool
            ),
            # pyre-ignore[6]: Expected `Union[None, bool, float, int, str]`
            #  for 8th param but got `Union[None, List[
            #  Union[None, bool, float, int, str]], bool, float, int, str]`.
            target_value=representation.get("target_value", None),
        )

    @classmethod
    def _make_choice_param(
        cls,
        name: str,
        representation: TParameterRepresentation,
        parameter_type: str | None,
    ) -> ChoiceParameter:
        values = representation["values"]
        assert isinstance(values, list) and len(values) > 1, (
            f"Cannot parse parameter {name}: for choice parameters, json representation"
            " should include a list of two or more values."
        )
        return ChoiceParameter(
            name=name,
            parameter_type=cls._to_parameter_type(
                values, parameter_type, name, "values"
            ),
            values=values,
            is_ordered=assert_is_instance_optional(
                representation.get("is_ordered"), bool
            ),
            is_fidelity=assert_is_instance(
                representation.get("is_fidelity", False), bool
            ),
            is_task=assert_is_instance(representation.get("is_task", False), bool),
            target_value=representation.get("target_value", None),  # pyre-ignore[6]
            sort_values=assert_is_instance_optional(
                representation.get("sort_values", None), bool
            ),
            dependents=assert_is_instance_optional(
                representation.get("dependents", None), dict
            ),
        )

    @classmethod
    def _make_fixed_param(
        cls,
        name: str,
        representation: TParameterRepresentation,
        parameter_type: str | None,
    ) -> FixedParameter:
        assert "value" in representation, "Value is required for fixed parameters."
        value = representation["value"]
        assert type(value) in PARAM_TYPES.values(), (
            f"Cannot parse fixed parameter {name}: for fixed parameters, json "
            "representation should include a single value."
        )
        return FixedParameter(
            name=name,
            parameter_type=(
                cls._get_parameter_type(type(value))  # pyre-ignore[6]
                if parameter_type is None
                # pyre-ignore[6]
                else cls._get_parameter_type(PARAM_TYPES[parameter_type])
            ),
            value=value,  # pyre-ignore[6]
            is_fidelity=assert_is_instance(
                representation.get("is_fidelity", False), bool
            ),
            target_value=representation.get("target_value", None),  # pyre-ignore[6]
            dependents=representation.get("dependents", None),  # pyre-ignore[6]
        )

    @classmethod
    def _make_derived_param(
        cls,
        name: str,
        representation: TParameterRepresentation,
        parameter_type: str | None,
    ) -> DerivedParameter:
        assert (
            "expression_str" in representation
        ), "expression_str is required for derived parameters."
        msg = "parameter_type is required for derived parameters."
        return DerivedParameter(
            name=name,
            parameter_type=cls._get_parameter_type(
                PARAM_TYPES[none_throws(parameter_type, msg)]  # pyre-ignore[6]
            ),
            expression_str=assert_is_instance(representation["expression_str"], str),
        )

    @classmethod
    def parameter_from_json(
        cls,
        representation: TParameterRepresentation,
    ) -> Parameter:
        """Instantiate a parameter from JSON representation."""
        if "parameter_type" in representation:
            raise ValueError(
                "'parameter_type' is not an expected key in parameter dictionary. "
                "If you are looking to specify the type of values that this "
                "parameter should take, use 'value_type' (expects 'int', 'float', "
                "'str' or 'bool')."
            )
        unexpected_keys = set(representation.keys()) - EXPECTED_KEYS_IN_PARAM_REPR
        if unexpected_keys:
            raise ValueError(
                f"Unexpected keys {unexpected_keys} in parameter representation."
                f"Exhaustive set of expected keys: {EXPECTED_KEYS_IN_PARAM_REPR}."
            )
        name = representation["name"]
        assert isinstance(name, str), "Parameter name must be a string."
        parameter_class = representation["type"]
        assert isinstance(parameter_class, str) and parameter_class in PARAM_CLASSES, (
            "Type in parameter JSON representation must be "
            "`range`, `choice`, `fixed` or `derived`."
        )

        parameter_type = representation.get("value_type", None)
        if parameter_type is not None:
            assert isinstance(parameter_type, str) and parameter_type in PARAM_TYPES, (
                "Value type in parameter JSON representation must be 'int', 'float', "
                "'bool' or 'str'."
            )

        if " " in name:
            raise ValueError(
                "Parameter names cannot contain spaces when used with AxClient. Got "
                f"{name!r}."
            )

        if parameter_class == "range":
            return cls._make_range_param(
                name=name,
                representation=representation,
                parameter_type=parameter_type,
            )

        if parameter_class == "choice":
            assert (
                "values" in representation
            ), "Values are required for choice parameters."
            values = representation["values"]
            if isinstance(values, list) and len(values) == 1:
                logger.info(
                    f"Choice parameter {name} contains only one value, converting to a"
                    + " fixed parameter instead."
                )
                # update the representation to a fixed parameter class
                parameter_class = "fixed"
                representation["type"] = parameter_class
                representation["value"] = values[0]
                del representation["values"]
            else:
                return cls._make_choice_param(
                    name=name,
                    representation=representation,
                    parameter_type=parameter_type,
                )

        if parameter_class == "fixed":
            assert not any(isinstance(val, list) for val in representation.values())
            return cls._make_fixed_param(
                name=name,
                representation=representation,
                parameter_type=parameter_type,
            )

        if parameter_class == "derived":
            return cls._make_derived_param(
                name=name,
                representation=representation,
                parameter_type=parameter_type,
            )

        raise ValueError(f"Unrecognized parameter type {parameter_class}.")

    @staticmethod
    def constraint_from_str(
        representation: str, parameters: dict[str, Parameter]
    ) -> ParameterConstraint:
        """Parse string representation of a parameter constraint.

        Args:
            representation: String representation of the parameter constraint.
            parameters: Dictionary of parameter names to parameter objects or configs.

        Returns:
            An instantiated ParameterConstraint, either an OrderConstraint or a
            ParameterConstraint, representing a linear constraint.
        """
        tokens = representation.split()
        try:
            float(tokens[-1])
            last_token_is_numeric = True
        except ValueError:
            last_token_is_numeric = False
        is_order_constraint = (
            len(tokens) == 3
            and tokens[1] in COMPARISON_OPS
            and not last_token_is_numeric
        )
        is_linear_constraint = (
            # if len == 3, then this is a single parameter bound constraint, otherwise
            # it corresponds to a numerical bound on a sum of parameters
            len(tokens) >= 3
            and len(tokens) % 2 == 1
            and tokens[-2] in COMPARISON_OPS
            and last_token_is_numeric
        )

        if is_order_constraint:  # e.g. "x1 >= x2"
            return _process_order_constraint(
                tokens=tokens,
                parameters=parameters,
            )

        if is_linear_constraint:  # e.g. "x1 + x2 >= 3"
            return _process_linear_constraint(
                tokens=tokens,
                parameters=parameters,
            )

        raise ValueError(INVALID_CONSTRAINT_ERROR_MSG.format(representation))

    @classmethod
    def outcome_constraint_from_str(
        cls,
        representation: str,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> OutcomeConstraint:
        """Parse string representation of an outcome constraint."""
        tokens = representation.split()
        assert len(tokens) == 3 and tokens[1] in COMPARISON_OPS, (
            f"Outcome constraint '{representation}' should be of "
            "form `metric_name >= x`, where x is a "
            "float bound and comparison operator is >= or <=."
        )
        op = COMPARISON_OPS[tokens[1]]
        rel = False
        try:
            bound_repr = tokens[2]
            if bound_repr[-1] == "%":
                rel = True
                bound_repr = bound_repr[:-1]
            bound = float(bound_repr)
        except ValueError:
            raise ValueError(
                f"Outcome constraint bound should be a float for '{representation}'."
            )
        return OutcomeConstraint(
            cls._make_metric(
                name=tokens[0],
                for_opt_config=True,
                metric_definitions=metric_definitions,
                lower_is_better=op is ComparisonOp.LEQ,
            ),
            op=op,
            bound=bound,
            relative=rel,
        )

    @classmethod
    def objective_threshold_constraint_from_str(
        cls,
        representation: str,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> ObjectiveThreshold:
        oc = cls.outcome_constraint_from_str(
            representation, metric_definitions=metric_definitions
        )
        return ObjectiveThreshold(
            metric=oc.metric.clone(),
            bound=oc.bound,
            relative=oc.relative,
            op=oc.op,
        )

    @classmethod
    def make_objectives(
        cls,
        objectives: dict[str, str],
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> list[Objective]:
        try:
            output_objectives = []
            for metric_name, min_or_max in objectives.items():
                minimize = (
                    MetricObjective[min_or_max.upper()] == MetricObjective.MINIMIZE
                )
                objective = Objective(
                    metric=cls._make_metric(
                        name=metric_name,
                        for_opt_config=True,
                        lower_is_better=minimize,
                        metric_definitions=metric_definitions,
                    ),
                    minimize=minimize,
                )
                output_objectives.append(objective)
            return output_objectives
        except KeyError as k:
            raise ValueError(
                "Objective values should specify "
                f"'{MetricObjective.MINIMIZE.name.lower()}' or "
                f"'{MetricObjective.MAXIMIZE.name.lower()}', got {k} in"
                f" objectives({objectives})"
            )

    @classmethod
    def make_outcome_constraints(
        cls,
        outcome_constraints: list[str],
        status_quo_defined: bool,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> list[OutcomeConstraint]:
        typed_outcome_constraints = [
            cls.outcome_constraint_from_str(c, metric_definitions=metric_definitions)
            for c in outcome_constraints
        ]

        if status_quo_defined is False and any(
            oc.relative for oc in typed_outcome_constraints
        ):
            raise ValueError(
                "Must set status_quo to have relative outcome constraints."
            )

        return typed_outcome_constraints

    @classmethod
    def make_objective_thresholds(
        cls,
        objective_thresholds: list[str],
        status_quo_defined: bool,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> list[ObjectiveThreshold]:
        typed_objective_thresholds = (
            [
                cls.objective_threshold_constraint_from_str(
                    c, metric_definitions=metric_definitions
                )
                for c in objective_thresholds
            ]
            if objective_thresholds is not None
            else []
        )

        if status_quo_defined is False and any(
            oc.relative for oc in typed_objective_thresholds
        ):
            raise ValueError(
                "Must set status_quo to have relative objective thresholds."
            )

        return typed_objective_thresholds

    @staticmethod
    def optimization_config_from_objectives(
        objectives: list[Objective],
        objective_thresholds: list[ObjectiveThreshold],
        outcome_constraints: list[OutcomeConstraint],
    ) -> OptimizationConfig:
        """Parse objectives and constraints to define optimization config.

        The resulting optimization config will be regular single-objective config
        if `objectives` is a list of one element and a multi-objective config
        otherwise.

        NOTE: If passing in multiple objectives, `objective_thresholds` must be a
        non-empty list defining constraints for each objective.
        """
        if len(objectives) == 1:
            if objective_thresholds:
                raise ValueError(
                    "Single-objective optimizations must not specify objective "
                    "thresholds."
                )
            return OptimizationConfig(
                objective=objectives[0],
                outcome_constraints=outcome_constraints,
            )

        if not objective_thresholds:
            logger.warning(
                "Objective thresholds were not set. They will be selected using a "
                "heuristic, but should be specified on the objective for best "
                "performance."
            )

        return MultiObjectiveOptimizationConfig(
            objective=MultiObjective(objectives=objectives),
            outcome_constraints=outcome_constraints,
            objective_thresholds=objective_thresholds,
        )

    @classmethod
    def make_optimization_config(
        cls,
        objectives: dict[str, str],
        objective_thresholds: list[str],
        outcome_constraints: list[str],
        status_quo_defined: bool,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
    ) -> OptimizationConfig:
        return cls.optimization_config_from_objectives(
            cls.make_objectives(objectives, metric_definitions=metric_definitions),
            cls.make_objective_thresholds(
                objective_thresholds,
                status_quo_defined,
                metric_definitions=metric_definitions,
            ),
            cls.make_outcome_constraints(
                outcome_constraints,
                status_quo_defined,
                metric_definitions=metric_definitions,
            ),
        )

    @classmethod
    def make_optimization_config_from_properties(
        cls,
        objectives: dict[str, ObjectiveProperties] | None = None,
        outcome_constraints: list[str] | None = None,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
        status_quo_defined: bool = False,
    ) -> OptimizationConfig | None:
        """Makes optimization config based on ObjectiveProperties objects

        Args:
            objectives: Mapping from an objective name to object containing:
                minimize: Whether this experiment represents a minimization problem.
                threshold: The bound in the objective's threshold constraint.
            outcome_constraints: List of string representation of outcome
                constraints of form "metric_name >= bound", like "m1 <= 3."
            status_quo_defined: bool for whether the experiment has a status quo
            metric_definitions: A mapping of metric names to extra kwargs to pass
                to that metric
        """
        if objectives is not None:
            objective_thresholds = (
                cls.build_objective_thresholds(objectives)
                if objectives is not None
                else []
            )
            simple_objectives = {
                objective: ("minimize" if properties.minimize else "maximize")
                for objective, properties in objectives.items()
            }
            return cls.make_optimization_config(
                objectives=simple_objectives,
                objective_thresholds=objective_thresholds,
                outcome_constraints=outcome_constraints or [],
                status_quo_defined=status_quo_defined,
                metric_definitions=metric_definitions,
            )
        return None

    @classmethod
    def make_search_space(
        cls,
        parameters: list[TParameterRepresentation],
        parameter_constraints: list[str] | None,
    ) -> SearchSpace:
        parameter_constraints = (
            parameter_constraints if parameter_constraints is not None else []
        )
        typed_parameters = [cls.parameter_from_json(p) for p in parameters]
        is_hss = any(p.is_hierarchical for p in typed_parameters)
        search_space_cls = HierarchicalSearchSpace if is_hss else SearchSpace

        parameter_map = {p.name: p for p in typed_parameters}

        typed_parameter_constraints = [
            cls.constraint_from_str(c, parameter_map) for c in parameter_constraints
        ]

        if any(
            any(
                isinstance(parameter_map[parameter], ChoiceParameter)
                for parameter in constraint.constraint_dict
            )
            for constraint in typed_parameter_constraints
        ):
            raise UnsupportedError(
                "Constraints on ChoiceParameters are not allowed. Try absorbing "
                "this constraint into the associated range parameter's bounds."
            )

        if any(
            any(
                isinstance(parameter_map[parameter], FixedParameter)
                for parameter in constraint.constraint_dict
            )
            for constraint in typed_parameter_constraints
        ):
            raise UnsupportedError(
                "Constraints on FixedParameters are not allowed. Try absorbing "
                "this constraint into the associated range parameter's bounds."
            )

        ss = search_space_cls(
            parameters=typed_parameters,
            parameter_constraints=typed_parameter_constraints,
        )

        logger.debug(f"Created search space: {ss}.")
        if is_hss:
            hss = assert_is_instance(ss, HierarchicalSearchSpace)
            logger.debug(
                "Hieararchical structure of the search space: \n"
                f"{hss.hierarchical_structure_str(parameter_names_only=True)}"
            )

        return search_space_cls(
            parameters=typed_parameters,
            parameter_constraints=typed_parameter_constraints,
        )

    @classmethod
    def _get_default_objectives(cls) -> dict[str, str] | None:
        """Get the default objective and its optimization direction.

        The return type is optional since some subclasses may not wish to
        use any optimization config by default.
        """
        return {Keys.DEFAULT_OBJECTIVE_NAME.value: "maximize"}

    @classmethod
    def make_experiment(
        cls,
        parameters: list[TParameterRepresentation],
        name: str | None = None,
        description: str | None = None,
        owners: list[str] | None = None,
        parameter_constraints: list[str] | None = None,
        outcome_constraints: list[str] | None = None,
        status_quo: TParameterization | None = None,
        experiment_type: str | None = None,
        tracking_metric_names: list[str] | None = None,
        metric_definitions: dict[str, dict[str, Any]] | None = None,
        objectives: dict[str, str] | None = None,
        objective_thresholds: list[str] | None = None,
        support_intermediate_data: bool = False,
        immutable_search_space_and_opt_config: bool = True,
        auxiliary_experiments_by_purpose: None
        | (dict[AuxiliaryExperimentPurpose, list[AuxiliaryExperiment]]) = None,
        default_trial_type: str | None = None,
        default_runner: Runner | None = None,
        is_test: bool = False,
    ) -> Experiment:
        """Instantiation wrapper that allows for Ax `Experiment` creation
        without importing or instantiating any Ax classes.

        Args:
            parameters: List of dictionaries representing parameters in the
                experiment search space.
                Required elements in the dictionaries are:
                1. "name" (name of parameter, string),
                2. "type" (type of parameter: "range", "fixed", "choice", or
                "derived", string),
                and one of the following:
                3a. "bounds" for range parameters (list of two values, lower bound
                first),
                3b. "values" for choice parameters (list of values), or
                3c. "value" for fixed parameters (single value).
                3d. "expression_str" for derived parameters (string containing
                the definition of the derived parameter).
                Optional elements are:
                1. "log_scale" (for float-valued range parameters, bool),
                2. "value_type" (to specify type that values of this parameter should
                take; expects "float", "int", "bool" or "str"),
                3. "is_fidelity" (bool) and "target_value" (float) for fidelity
                parameters,
                4. "is_ordered" (bool) for choice parameters,
                5. "is_task" (bool) for task parameters, and
                6. "digits" (int) for float-valued range parameters.
            name: Name of the experiment to be created.
            parameter_constraints: List of string representation of parameter
                constraints, such as "x3 >= x4" or "-x3 + 2*x4 - 3.5*x5 >= 2". For
                the latter constraints, any number of arguments is accepted, and
                acceptable operators are "<=" and ">=".
            outcome_constraints: List of string representation of outcome
                constraints of form "metric_name >= bound", like "m1 <= 3."
            status_quo: Parameterization of the current state of the system.
                If set, this will be added to each trial to be evaluated alongside
                test configurations.
            experiment_type: String indicating type of the experiment (e.g. name of
                a product in which it is used), if any.
            tracking_metric_names: Names of additional tracking metrics not used for
                optimization.
            metric_definitions: A mapping of metric names to extra kwargs to pass
                to that metric
            objectives: Mapping from an objective name to "minimize" or "maximize"
                representing the direction for that objective.
            objective_thresholds: A list of objective threshold constraints for multi-
                objective optimization, in the same string format as
                `outcome_constraints` argument.
            support_intermediate_data: Whether trials may report metrics results for
                incomplete runs.
            immutable_search_space_and_opt_config: Whether it's possible to update the
                search space and optimization config on this experiment after creation.
                Defaults to True. If set to True, we won't store or load copies of the
                search space and optimization config on each generator run, which will
                improve storage performance.
            auxiliary_experiments_by_purpose: Dictionary of auxiliary experiments for
                different use cases (e.g., transfer learning).
            default_trial_type: The default trial type if multiple
                trial types are intended to be used in the experiment.  If specified,
                a MultiTypeExperiment will be created. Otherwise, a single-type
                Experiment will be created.
            default_runner: The default runner in this experiment.
                This only applies to MultiTypeExperiment (when default_trial_type
                is specified).
            is_test: Whether this experiment will be a test experiment (useful for
                marking test experiments in storage etc). Defaults to False.

        """
        if (default_trial_type is None) != (default_runner is None):
            raise ValueError(
                "Must specify both default_trial_type and default_runner if "
                "using a MultiTypeExperiment."
            )

        status_quo_arm = None if status_quo is None else Arm(parameters=status_quo)

        objectives = objectives or cls._get_default_objectives()
        if objectives:
            optimization_config = cls.make_optimization_config(
                objectives=objectives,
                objective_thresholds=objective_thresholds or [],
                outcome_constraints=outcome_constraints or [],
                status_quo_defined=status_quo_arm is not None,
                metric_definitions=metric_definitions,
            )
        else:
            optimization_config = None

        tracking_metrics = (
            None
            if tracking_metric_names is None
            else [
                cls._make_metric(
                    name=metric_name, metric_definitions=metric_definitions
                )
                for metric_name in tracking_metric_names
            ]
        )

        default_data_type = (
            DataType.MAP_DATA if support_intermediate_data else DataType.DATA
        )

        properties: dict[str, Any] = {}

        if immutable_search_space_and_opt_config:
            properties[Keys.IMMUTABLE_SEARCH_SPACE_AND_OPT_CONF] = (
                immutable_search_space_and_opt_config
            )

        if owners is not None:
            properties["owners"] = owners
        if default_trial_type is not None:
            return MultiTypeExperiment(
                name=none_throws(name),
                search_space=cls.make_search_space(
                    parameters=parameters, parameter_constraints=parameter_constraints
                ),
                default_trial_type=none_throws(default_trial_type),
                default_runner=none_throws(default_runner),
                optimization_config=optimization_config,
                tracking_metrics=tracking_metrics,
                status_quo=status_quo_arm,
                description=description,
                is_test=is_test,
                experiment_type=experiment_type,
                properties=properties,
                default_data_type=default_data_type,
            )

        return Experiment(
            name=name,
            description=description,
            search_space=cls.make_search_space(parameters, parameter_constraints),
            optimization_config=optimization_config,
            status_quo=status_quo_arm,
            experiment_type=experiment_type,
            tracking_metrics=tracking_metrics,
            default_data_type=default_data_type,
            properties=properties,
            auxiliary_experiments_by_purpose=auxiliary_experiments_by_purpose,
            is_test=is_test,
        )

    @classmethod
    def build_objective_thresholds(
        cls, objectives: dict[str, ObjectiveProperties]
    ) -> list[str]:
        """Construct a list of constraint string for an objective thresholds
        interpretable by `make_experiment()`

        Args:
            objectives: Mapping of name of the objective to Object containing:
                minimize: Whether this experiment represents a minimization problem.
                threshold: The bound in the objective's threshold constraint.
        """
        return [
            cls.build_objective_threshold(objective, properties)
            for objective, properties in objectives.items()
            if properties.threshold is not None
        ]

    @staticmethod
    def build_objective_threshold(
        objective: str, objective_properties: ObjectiveProperties
    ) -> str:
        """
        Constructs constraint string for an objective threshold interpretable
        by `make_experiment()`

        Args:
            objective: Name of the objective
            objective_properties: Object containing:
                minimize: Whether this experiment represents a minimization problem.
                threshold: The bound in the objective's threshold constraint.
        """
        operator = "<=" if objective_properties.minimize else ">="
        return f"{objective} {operator} {objective_properties.threshold}"

    @staticmethod
    def make_fixed_observation_features(
        fixed_features: FixedFeatures,
    ) -> ObservationFeatures:
        """Construct ObservationFeatures from FixedFeatures.

        Args:
            fixed_features: The fixed features for generation.

        Returns:
            The new ObservationFeatures object.
        """
        return ObservationFeatures(
            parameters=fixed_features.parameters,
            trial_index=(
                None
                if fixed_features.trial_index is None
                else fixed_features.trial_index
            ),
        )


# Helpers for parsing parameter constraints
def _process_order_constraint(
    tokens: Sequence[str],
    parameters: Mapping[str, Parameter],
) -> OrderConstraint:
    """Processes an order constraint, e.g. "x1 <= x2".

    Args:
        tokens: A list of tokens in the constraint string.
        parameters: A mapping from parameter names to their definitions.

    Returns:
        An OrderConstraint object representing the order constraint.
    """
    left, right = tokens[0], tokens[2]
    parameter_names = parameters.keys()
    if missing_tokens := ({left, right} - set(parameter_names)):
        missing_tokens = list(missing_tokens)
        missing_tokens.sort()  # making error message deterministic
        raise ValueError(
            f"OrderConstraint token(s) {missing_tokens} are not present in "
            f"parameters list {list(parameter_names)}."
        )

    validate_constraint_parameters(parameters=[parameters[left], parameters[right]])
    # tokens[1] is checked to be either LEQ or GEQ above if order_const is True
    if COMPARISON_OPS[tokens[1]] is ComparisonOp.LEQ:
        lower_parameter = parameters[left]
        upper_parameter = parameters[right]
    else:
        lower_parameter = parameters[right]
        upper_parameter = parameters[left]
    return OrderConstraint(
        lower_parameter=lower_parameter, upper_parameter=upper_parameter
    )


def _process_linear_constraint(
    tokens: Sequence[str],
    parameters: Mapping[str, Parameter],
) -> ParameterConstraint:
    """Processes a linear constraint, e.g. "x1 + x2 <= 3". The last token is expected
    to be a numeric constant, and the other tokens are expected to be parameters, their
    multiplicative coefficients (e.g."2.5*x1") and "+" or "-" operators (e.g. "+").

    Args:
        tokens: A list of tokens in the constraint string.
        parameters: A mapping from parameter names to their definitions.

    Returns:
        A ParameterConstraint object representing the linear constraint.
    """
    parameter_names = parameters.keys()

    bound = float(tokens[-1])
    if any(token[0] == "*" or token[-1] == "*" for token in tokens):
        raise ValueError(
            "A linear constraint should be the form a*x + b*y - c*z <= d"
            ", where a,b,c,d are float constants and x,y,z are parameters. "
            "There should be no space in each term around the operator `*`, and "
            "there should be a single space around each operator +, -, <= and >=."
        )
    parameter_weights = {}
    current_sign = 1.0  # Determines whether the operator is + or -
    # tokens are alternating monomials and operators
    for idx, token in enumerate(tokens[:-2]):
        # for monomials
        if idx % 2 == 0:
            multiplier, parameter_name = _process_monomial(monomial_str=token)
            if parameter_name not in parameter_names:
                raise ValueError(
                    f"Constraint parameter '{parameter_name}' is not present in this "
                    f"experiment's search space parameters: {list(parameter_names)}."
                )
            validate_constraint_parameters(parameters=[parameters[parameter_name]])
            parameter_weights[parameter_name] = current_sign * multiplier
        # for operators
        else:
            if token == "+":
                current_sign = 1.0
            elif token == "-":
                current_sign = -1.0
            else:
                raise ValueError(
                    f"Expected a mixed constraint, found operator `{token}`."
                )
    # tokens[-2] is checked to be either LEQ or GEQ if sum_const is True
    comparison_multiplier = (
        1.0 if COMPARISON_OPS[tokens[-2]] is ComparisonOp.LEQ else -1.0
    )
    return ParameterConstraint(
        constraint_dict={
            p: comparison_multiplier * parameter_weights[p] for p in parameter_weights
        },
        bound=comparison_multiplier * bound,
    )


def _process_monomial(monomial_str: str) -> tuple[float, str]:
    """Process a monomial in a linear constraint.

    Args:
        monomial_str: A string representation of a monomial in a linear constraint.

    Returns:
        A tuple of the multiplier and the parameter name.
    """
    split_token = monomial_str.split("*")
    parameter = ""  # Initializing the parameter
    multiplier = 1.0  # Initializing the multiplier
    if len(split_token) == 2:  # There is a non-unit multiplier
        try:
            multiplier = float(split_token[0])
        except ValueError:
            raise ValueError(
                f"Multiplier should be float; got {split_token[0]} for "
                f"parameter {split_token[1]}."
            )
        parameter = split_token[1]
    elif len(split_token) == 1:  # The multiplier is either -1 or 1
        parameter = split_token[0]
        if parameter[0] == "-":  # The multiplier is -1
            parameter = parameter[1:]
            multiplier = -1.0
        else:
            multiplier = 1.0
    else:
        raise ValueError(
            "Monomial format does not match `multiplier*parameter_name`."
            f"Got `{monomial_str}`."
        )
    return multiplier, parameter



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import inspect
from collections import Counter
from datetime import datetime
from typing import Any, get_type_hints

from ax.adapter.registry import Generators

from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.exceptions.generation_strategy import AxGenerationException
from ax.generation_strategy.generation_node import GenerationNode
from ax.generation_strategy.generation_node_input_constructors import (
    InputConstructorPurpose,
    NodeInputConstructors,
)
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.generation_strategy.generator_spec import GeneratorSpec
from ax.utils.common.constants import Keys
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment

# type hints are stored as str when using from __future__ import annotations
EXPECTED_INPUT_CONSTRUCTOR_PARAMETER_ANNOTATIONS = [
    inspect.Parameter(
        name="previous_node",
        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        annotation="GenerationNode | None",
    ),
    inspect.Parameter(
        name="next_node",
        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        annotation="GenerationNode",
    ),
    inspect.Parameter(
        name="gs_gen_call_kwargs",
        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        annotation="dict[str, Any]",
    ),
    inspect.Parameter(
        name="experiment",
        kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
        annotation="Experiment",
    ),
]


class TestGenerationNodeInputConstructors(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.sobol_generator_spec = GeneratorSpec(
            generator_enum=Generators.SOBOL,
            model_kwargs={"init_position": 3},
            model_gen_kwargs={"some_gen_kwarg": "some_value"},
        )
        self.sobol_generation_node = GenerationNode(
            node_name="test", generator_specs=[self.sobol_generator_spec]
        )
        self.experiment = get_branin_experiment()
        # construct a list of grs that will mock a list of grs that would exist during
        # a gs.gen call. This list has one single arm GR, and one 3-arm GR.
        self.grs = [
            GeneratorRun(arms=[Arm(parameters={"x1": 1, "x2": 5})]),
            GeneratorRun(arms=[Arm(parameters={"x1": 1, "x2": y}) for y in range(3)]),
        ]
        self.gs = GenerationStrategy(nodes=[self.sobol_generation_node], name="test")
        # NOTE: This mapping relies on names of the NodeInputConstructors members
        # aligning with the names of the InputConstructorPurpose members, like so:
        # InputConstructorPurpose.N matches NodeInputConstructors.*_N. We may need
        # to switch to constructing this mapping manually in the future.
        self.purposes_to_input_constructors = {
            p: [ip for ip in NodeInputConstructors if ip.name.endswith(f"{p.name}")]
            for p in InputConstructorPurpose
        }

        # type hints are stored as str when using from __future__ import annotations
        self.all_purposes_expected_signatures = {
            InputConstructorPurpose.N: inspect.Signature(
                parameters=EXPECTED_INPUT_CONSTRUCTOR_PARAMETER_ANNOTATIONS,
                return_annotation="int",
            ),
            InputConstructorPurpose.FIXED_FEATURES: inspect.Signature(
                parameters=EXPECTED_INPUT_CONSTRUCTOR_PARAMETER_ANNOTATIONS,
                return_annotation="ObservationFeatures | None",
            ),
        }

    def test_all_constructors_have_expected_signature_for_purpose(self) -> None:
        """Test that all node input constructors methods have the same signature
        and that the parameters are of the expected types."""
        untested_constructors = set(NodeInputConstructors)
        for purpose, constructors in self.purposes_to_input_constructors.items():
            # For each purpose, we check that all constructors that match it,
            # share the same expected signature.
            with self.subTest(purpose=purpose, constructors=constructors):
                self.assertIn(purpose, self.all_purposes_expected_signatures)
                for constructor in constructors:
                    self.assertEqual(
                        inspect.signature(constructor._get_function_for_value()),
                        self.all_purposes_expected_signatures[purpose],
                    )
                    untested_constructors.remove(constructor)

        # There should be no untested constructors left.
        self.assertEqual(len(untested_constructors), 0)

    def test_consume_all_n_constructor(self) -> None:
        """Test that the consume_all_n_constructor returns full n."""
        num_to_gen = NodeInputConstructors.ALL_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 5},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 5)

    def test_repeat_arm_n_constructor(self) -> None:
        """Test that the repeat_arm_n_constructor returns a small percentage of n."""
        medium_n = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 8},
            experiment=self.experiment,
        )
        large_n = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 11},
            experiment=self.experiment,
        )
        self.assertEqual(medium_n, 1)
        self.assertEqual(large_n, 2)

    def test_repeat_arm_n_constructor_return_0(self) -> None:
        small_n = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 5},
            experiment=self.experiment,
        )
        self.assertEqual(small_n, 0)
        self.assertTrue(self.sobol_generation_node._should_skip)

    def test_remaining_n_constructor_expect_1(self) -> None:
        """Test that the remaining_n_constructor returns the remaining n."""
        # should return 1 because 4 arms already exist and 5 are requested
        expect_1 = NodeInputConstructors.REMAINING_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 5, "grs_this_gen": self.grs},
            experiment=self.experiment,
        )
        self.assertEqual(expect_1, 1)

    def test_remaining_n_constructor_expect_0(self) -> None:
        # should return 0 because 4 arms already exist and 4 are requested
        expect_0 = NodeInputConstructors.REMAINING_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 4, "grs_this_gen": self.grs},
            experiment=self.experiment,
        )
        self.assertEqual(expect_0, 0)

    def test_remaining_n_constructor_cap_at_zero(self) -> None:
        # should return 0 because 4 arms already exist and 3 are requested
        # this is a bad state that should never be hit, but ensuring proper
        # handling here feels like a valid edge case
        expect_0 = NodeInputConstructors.REMAINING_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"n": 3, "grs_this_gen": self.grs},
            experiment=self.experiment,
        )
        self.assertEqual(expect_0, 0)

    def test_no_n_provided_all_n(self) -> None:
        num_to_gen = NodeInputConstructors.ALL_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 10)

    def test_no_n_provided_all_n_with_exp_prop(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 12
        num_to_gen = NodeInputConstructors.ALL_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 12)

    def test_no_n_provided_all_n_with_exp_prop_long_run(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 13
        self.sobol_generation_node._trial_type = Keys.LONG_RUN
        num_to_gen = NodeInputConstructors.ALL_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 7)

    def test_no_n_provided_all_n_with_exp_prop_short_run(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 13
        self.sobol_generation_node._trial_type = Keys.SHORT_RUN
        num_to_gen = NodeInputConstructors.ALL_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 6)

    def test_no_n_provided_repeat_n(self) -> None:
        num_to_gen = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 1)

    def test_no_n_provided_repeat_n_with_exp_prop(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 18
        num_to_gen = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 2)

    def test_no_n_provided_repeat_n_with_exp_prop_long_run(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 18
        self.sobol_generation_node._trial_type = Keys.SHORT_RUN
        num_to_gen = NodeInputConstructors.REPEAT_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        # expect 1 arm here because total concurrent arms is 18, and we have a trial
        # type (short run), so we'll take the floor of 18/2 = 9 to be used in the
        # logic for repeat arms which says if we have less than 10 requested arms we
        # should get 1 repeat arm.
        self.assertEqual(num_to_gen, 1)

    def test_no_n_provided_remaining_n(self) -> None:
        num_to_gen = NodeInputConstructors.REMAINING_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 10)

    def test_no_n_provided_remaining_n_with_exp_prop(self) -> None:
        self.experiment._properties[Keys.EXPERIMENT_TOTAL_CONCURRENT_ARMS] = 8
        num_to_gen = NodeInputConstructors.REMAINING_N(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={"grs_this_gen": self.grs},
            experiment=self.experiment,
        )
        self.assertEqual(num_to_gen, 4)

    def test_set_target_trial_long_run_wins(self) -> None:
        for num_arms, trial_type in zip((1, 3), (Keys.LONG_RUN, Keys.SHORT_RUN)):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=trial_type,
                complete=False,
                num_arms=num_arms,
            )
        self.experiment.fetch_data()
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=0,
            ),
        )

    def test_set_target_trial_most_arms_long_run_wins(self) -> None:
        for num_arms in (1, 3):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.LONG_RUN,
                complete=False,
                num_arms=num_arms,
            )
        self.experiment.fetch_data()
        # Test most arms should win
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_long_run_ties(self) -> None:
        # if all things are equal we should just pick the first one
        # in the sorted list
        for _ in range(2):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.LONG_RUN,
                complete=False,
                num_arms=1,
            )
        self.experiment.fetch_data()
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=0,
            ),
        )

    def test_set_target_trial_longest_duration_long_run_wins(self) -> None:
        for _ in range(2):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.LONG_RUN,
                complete=False,
                num_arms=1,
            )
        self.experiment.fetch_data()
        self.experiment.trials[0]._time_run_started = datetime(2000, 1, 2)
        self.experiment.trials[1]._time_run_started = datetime(2000, 1, 1)
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_running_short_trial_wins(self) -> None:
        for trial_type, complete in zip((Keys.LONG_RUN, Keys.SHORT_RUN), (True, False)):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=trial_type,
                complete=complete,
                num_arms=1,
            )
        self.experiment.fetch_data()
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_longest_short_wins(self) -> None:
        for _ in range(2):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.SHORT_RUN,
                complete=False,
                num_arms=1,
            )
        self.experiment.fetch_data()
        self.experiment.trials[0]._time_run_started = datetime(2000, 1, 2)
        self.experiment.trials[1]._time_run_started = datetime(2000, 1, 1)
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_most_arms_short_running_wins(self) -> None:
        for num_arms in (1, 3):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.SHORT_RUN,
                complete=False,
                num_arms=num_arms,
            )
        self.experiment.fetch_data()
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_most_arms_complete_short_wins(self) -> None:
        for num_arms in (1, 3):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.SHORT_RUN,
                complete=False,
                num_arms=num_arms,
            )
        self.experiment.fetch_data()
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_set_target_trial_longest_short_complete_wins(self) -> None:
        for _ in range(2):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.SHORT_RUN,
                complete=True,
                num_arms=1,
            )
        self.experiment.fetch_data()
        self.experiment.trials[0]._time_run_started = datetime(2000, 1, 2)
        self.experiment.trials[1]._time_run_started = datetime(2000, 1, 1)
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=1,
            ),
        )

    def test_target_trial_raises_error_if_none_found(self) -> None:
        with self.assertRaisesRegex(
            AxGenerationException,
            "Often this could be due to no trials",
        ):
            NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
                previous_node=None,
                next_node=self.sobol_generation_node,
                gs_gen_call_kwargs={},
                experiment=self.experiment,
            )

    def test_target_trial_latest_trial_no_data(self) -> None:
        for _ in range(2):
            self._add_sobol_trial(
                experiment=self.experiment,
                trial_type=Keys.SHORT_RUN,
                complete=True,
                num_arms=1,
            )
        self.experiment.fetch_data(trial_indices=[0])
        target_trial = NodeInputConstructors.TARGET_TRIAL_FIXED_FEATURES(
            previous_node=None,
            next_node=self.sobol_generation_node,
            gs_gen_call_kwargs={},
            experiment=self.experiment,
        )
        self.assertEqual(
            target_trial,
            ObservationFeatures(
                parameters={},
                trial_index=0,
            ),
        )

    def _add_sobol_trial(
        self,
        experiment: Experiment,
        trial_type: str | None = None,
        complete: bool = True,
        num_arms: int = 1,
        with_status_quo: bool = True,
    ) -> BatchTrial:
        """Helper function to add a trial to an experiment, takes a trial type and
        whether or not the trial is complete, and number of arms"""
        grs = []
        for i in range(num_arms):
            grs.append(GeneratorRun(arms=[Arm(parameters={"x1": 1, "x2": i})]))
        trial = experiment.new_batch_trial(
            trial_type=trial_type,
            generator_runs=grs,
        )
        if with_status_quo:
            if experiment.status_quo is None:
                experiment.status_quo = Arm(parameters={"x1": 0, "x2": 0})
            trial.add_status_quo_arm(
                weight=1.0,
            )
        trial.run()
        if complete:
            trial.mark_completed()
        return trial


class TestInstantiationFromNodeInputConstructor(TestCase):
    """Class to test that all node input constructors can be instantiated and are
    being tested."""

    def setUp(self) -> None:
        super().setUp()
        self.constructor_cases = {
            "ALl_N": NodeInputConstructors.ALL_N,
            "REPEAT_N": NodeInputConstructors.REPEAT_N,
            "REMAINING_N": NodeInputConstructors.REMAINING_N,
        }
        self.purpose_cases = {
            "N": InputConstructorPurpose.N,
        }

    def test_all_constructors_have_same_signature(self) -> None:
        """Test that all node input constructors methods have the same signature
        and that the parameters are of the expected types"""
        all_constructors_tested = list(self.constructor_cases.values())
        method_signature = inspect.signature(all_constructors_tested[0])
        for constructor in all_constructors_tested[1:]:
            with self.subTest(constructor=constructor):
                func_parameters = get_type_hints(constructor.__call__)
                self.assertEqual(
                    Counter(list(func_parameters.keys())),
                    Counter(
                        [
                            "previous_node",
                            "next_node",
                            "gs_gen_call_kwargs",
                            "experiment",
                            "return",
                        ]
                    ),
                )
                self.assertEqual(
                    func_parameters["previous_node"], GenerationNode | None
                )
                self.assertEqual(func_parameters["next_node"], GenerationNode)
                self.assertEqual(func_parameters["gs_gen_call_kwargs"], dict[str, Any])
                self.assertEqual(func_parameters["experiment"], Experiment)
                self.assertEqual(method_signature, inspect.signature(constructor))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable
from typing import Any

import numpy as np
import torch
from ax.core import Arm, GeneratorRun, OptimizationConfig
from ax.core.experiment import Experiment
from ax.core.types import TEvaluationOutcome, TParameterization
from ax.service.utils.instantiation import InstantiationBase
from ax.utils.common.constants import Keys
from botorch.utils.sampling import draw_sobol_samples
from pyre_extensions import assert_is_instance, none_throws

# from ExperimentType in ae/lazarus/fb/utils/if/ae.thrift
PBO_EXPERIMENT_TYPE: str = "PREFERENCE_LEARNING"
PE_EXPERIMENT_TYPE: str = "PREFERENCE_EXPLORATION"


def sum_utility(parameters: TParameterization) -> float:
    """Test utility function that sums over parameter values"""
    values = [assert_is_instance(v, float) for v in parameters.values()]
    return sum(values)


def pairwise_pref_metric_eval(
    parameters: dict[str, TParameterization],
    utility_func: Callable[[TParameterization], float] = sum_utility,
) -> dict[str, TEvaluationOutcome]:
    """evaluating pairwise comparisons using utility_func"""
    assert len(parameters.keys()) == 2
    arm1, arm2 = list(parameters.keys())
    arm1_sum, arm2_sum = sum_utility(parameters[arm1]), sum_utility(parameters[arm2])
    is_arm1_preferred = int(arm1_sum - arm2_sum > 0)
    return {
        arm1: {Keys.PAIRWISE_PREFERENCE_QUERY.value: is_arm1_preferred},
        arm2: {Keys.PAIRWISE_PREFERENCE_QUERY.value: 1 - is_arm1_preferred},
    }


def _metric_name_to_value(metric_name: str, metric_names: list[str]) -> float:
    i = sorted(metric_names).index(metric_name)
    return (i + 1) * 1000.0 + np.random.standard_normal()


def experimental_metric_eval(
    parameters: dict[str, Any], metric_names: list[str]
) -> dict[str, TEvaluationOutcome]:
    """evaluating experimental metrics

    Args:
        parameters: Dict of arm name to parameterization
        metric_names: List of metric names

    Returns:
        Dict of arm name to metric name to (mean, sem)
    """
    result_dict = {}
    for arm_name in parameters.keys():
        result_dict[arm_name] = {
            metric_name: (
                _metric_name_to_value(metric_name, metric_names),
                10.0,
            )
            for metric_name in metric_names
        }
    return result_dict


def get_pbo_experiment(
    num_parameters: int = 2,
    num_experimental_metrics: int = 3,
    parameter_names: list[str] | None = None,
    tracking_metric_names: list[str] | None = None,
    num_experimental_trials: int = 3,
    num_preference_trials: int = 3,
    num_preference_trials_w_repeated_arm: int = 5,
    include_sq: bool = True,
    partial_data: bool = False,
    unbounded_search_space: bool = False,
    experiment_name: str = "pref_experiment",
    optimization_config: OptimizationConfig | None = None,
) -> Experiment:
    """Create synthetic preferential BO experiment"""
    if tracking_metric_names is None:
        tracking_metric_names = [
            f"metric{i}" for i in range(1, num_experimental_metrics + 1)
        ]
    else:
        assert len(tracking_metric_names) == num_experimental_metrics

    if parameter_names is None:
        parameter_names = [f"x{i}" for i in range(1, num_parameters + 1)]
    else:
        assert len(parameter_names) == num_parameters

    param_bounds = [10.0, 30.0] if not unbounded_search_space else [-1e9, 1e9]
    if include_sq:
        sq = {param_name: np.mean(param_bounds) for param_name in parameter_names}
        sq_arm = Arm(parameters=sq, name="status_quo")
    else:
        sq = None
        sq_arm = None

    parameters = [
        {
            "name": param_name,
            "type": "range",
            # make the default search space non-unit for better clarity in testing
            "bounds": param_bounds,  # Changed from list to tuple
        }
        for param_name in parameter_names
    ]

    has_preference_query = (
        num_preference_trials > 0 or num_preference_trials_w_repeated_arm > 0
    )

    if optimization_config:
        opt_config = optimization_config
    elif has_preference_query:
        opt_config = InstantiationBase.make_optimization_config(
            objectives={Keys.PAIRWISE_PREFERENCE_QUERY.value: "maximize"},
            objective_thresholds=[],
            outcome_constraints=[],
            status_quo_defined=True,
        )
    elif len(tracking_metric_names) > 0:
        opt_config = InstantiationBase.make_optimization_config(
            objectives={tracking_metric_names[0]: "maximize"},
            objective_thresholds=[],
            outcome_constraints=[],
            status_quo_defined=True,
        )
    else:
        opt_config = None

    tracking_metrics = [
        InstantiationBase._make_metric(name=metric_name)
        for metric_name in tracking_metric_names
    ]

    experiment = Experiment(
        name=experiment_name,
        description="This is a test exp",
        experiment_type="NOTEBOOK",
        search_space=InstantiationBase.make_search_space(
            # pyre-fixme[6]: Incompatible parameter type
            parameters=parameters,
            parameter_constraints=None,
        ),
        optimization_config=opt_config,
        tracking_metrics=tracking_metrics,
        is_test=True,
        status_quo=sq_arm,
    )

    # Adding arms with experimental metrics
    t_bounds = torch.full(
        (2, len(parameter_names)), param_bounds[0], dtype=torch.double
    )
    t_bounds[1] = param_bounds[1]
    X = None
    if num_experimental_trials > 0:
        X = draw_sobol_samples(
            bounds=t_bounds, n=num_experimental_trials, q=1, seed=0
        ).squeeze(1)
    for t in range(num_experimental_trials):
        arm = {}
        for i, param_name in enumerate(experiment.search_space.parameters.keys()):
            arm[param_name] = none_throws(X)[t, i].item()
        gr = (
            # pyre-ignore: Incompatible parameter type [6]
            GeneratorRun([Arm(arm), Arm(sq)])
            if include_sq
            else GeneratorRun([Arm(arm)])
        )
        trial = experiment.new_batch_trial(generator_run=gr)
        raw_data = experimental_metric_eval(
            parameters={a.name: a.parameters for a in trial.arms},
            metric_names=tracking_metric_names,
        )
        # create incomplete data by dropping the first metric
        if partial_data:
            for v in raw_data.values():
                del assert_is_instance(v, dict)[tracking_metric_names[-1]]
        trial.attach_batch_trial_data(raw_data=raw_data)
        trial.mark_running(no_runner_required=True)
        trial.mark_completed()

    # Adding arms with preferential queries
    if not unbounded_search_space and num_preference_trials > 0:
        X = draw_sobol_samples(
            bounds=t_bounds, n=2 * num_preference_trials, q=1, seed=0
        ).squeeze(1)
    for t in range(num_preference_trials):
        arms = []
        for j in range(2):
            param_dict = {}
            for i, param_name in enumerate(experiment.search_space.parameters.keys()):
                # if this experiment used as PE experiment
                if unbounded_search_space:
                    # matching how metrics are generated in experimental_metric_eval
                    param_dict[param_name] = _metric_name_to_value(
                        metric_name=param_name, metric_names=parameter_names
                    )
                else:
                    param_dict[param_name] = none_throws(X)[t * 2 + j, i].item()
            arms.append(Arm(parameters=param_dict))
        gr = GeneratorRun(arms)

        trial = experiment.new_batch_trial(generator_run=gr)
        trial.attach_batch_trial_data(
            raw_data=pairwise_pref_metric_eval(
                parameters={a.name: a.parameters for a in trial.arms}
            )
        )
        trial.mark_running(no_runner_required=True)
        trial.mark_completed()

    # Adding preferential queries using previously evaluated arms
    for _ in range(num_preference_trials_w_repeated_arm):
        arms = np.random.choice(
            list(experiment.arms_by_name.values()), 2, replace=False
        )
        trial = experiment.new_batch_trial()
        trial.add_arms_and_weights(arms=arms)
        trial.attach_batch_trial_data(
            raw_data=pairwise_pref_metric_eval(
                parameters={a.name: a.parameters for a in trial.arms}
            )
        )
        trial.mark_running(no_runner_required=True)
        trial.mark_completed()

    return experiment
